import torch
import torch.nn as nn
import torch.nn.functional as F

def conv(ic:int, oc:int, ks:int=3, s:int=1, p:int=0, norm:bool=False, act:bool=True):
    layers = [nn.Conv2d(ic, oc, kernel_size=ks, stride=s, padding=p, bias=not norm)]
    if norm: layers += [nn.BatchNorm2d(oc)]
    if act: layers += [nn.ReLU(inplace=True)]
    return layers

def mpool(ks:int=2, s:int=2, p:int=0):
    return nn.MaxPool2d(kernel_size=ks, stride=s, padding=p)

def uconv(ic:int, oc:int, ks:int=2, s:int=2, p:int=0, norm:bool=False, act:bool=False):
    layers = [nn.ConvTranspose2d(ic, oc, kernel_size=ks, stride=s, padding=p, bias=not norm)]
    if norm: layers += [nn.BatchNorm2d(oc)]
    if act: layers += [nn.ReLU(inplace=True)]
    return layers

class UNet_EBlock(nn.Module):
    def __init__(self, ic):
        super().__init__()
        self.uc = uconv(ic, ic//2)[0]
        self.up = nn.Sequential(*conv(ic, ic//2), *conv(ic//2, ic//2))

    def forward(self, i1, i2):
        '''
        Arguments:
        :: i1 - output of the layer below
        :: i2 - output from the side layer
        '''
        dc = self.uc(i1)
        h1, h2, w1, w2 = i2.shape[2], dc.shape[2], i2.shape[3], dc.shape[3]
        d1, d2 = (h1 - h2) // 2, (w1 - w2) // 2
        i2 = i2[:, :, d1:d1+h2, d2:d2+w2]
        return self.up(torch.cat((dc, i2), dim=1))
        
class UNet(nn.Module):
    '''
    The UNet Architecture.
    Reproduced from paper: https://arxiv.org/pdf/1505.04597.pdf

    Arguments:
    :: ic - # of input channels
    :: nc - # of classes
    :: sd - # of filters for first convolution
    :: sos - If True, output is of the same size as the input
             Default: True
    '''

    def __init__(self, ic, nc, sd=64, sos=True):
        
        super().__init__()
        
        self.ic = ic; self.nc = nc; self.sos = sos
        
        self.ud1 = nn.Sequential(*conv(ic, sd), *conv(sd, sd))
        self.ud2 = nn.Sequential(mpool(), *conv(sd, sd*2), *conv(sd*2, sd*2))
        self.ud3 = nn.Sequential(mpool(), *conv(sd*2, sd*4), *conv(sd*4, sd*4))
        self.ud4 = nn.Sequential(mpool(), *conv(sd*4, sd*8), *conv(sd*8, sd*8))
        self.ud5 = nn.Sequential(mpool(), *conv(sd*8, sd*16), *conv(sd*16, sd*16))

        self.ue1 = UNet_EBlock(sd*16)
        self.ue2 = UNet_EBlock(sd*8)
        self.ue3 = UNet_EBlock(sd*4)
        self.ue4 = UNet_EBlock(sd*2)

        self.fconv = conv(sd, nc, ks=1, act=False)[0]

    def forward(self, x):
        _, _, h, w = x.shape
        o1 = self.ud1(x); o2 = self.ud2(o1)
        o3 = self.ud3(o2); o4 = self.ud4(o3)
        o5 = self.ud5(o4)

        o6 = self.ue1(o5, o4.clone()); o7 = self.ue2(o6, o3.clone())
        o8 = self.ue3(o7, o2.clone()); o9 = self.ue4(o8, o1.clone())
        
        if self.sos:
            return F.interpolate(self.fconv(o9), (h, w), mode='bilinear', align_corners=False)
        else:
            return self.fconv(o9)



import numpy as np
import os
from PIL import Image

class DatasetLoader(object):

    def __init__(self, input_path, label_path):
        '''
        This is the Base DatasetLoader class

        INPUTS:
        - input_path : The path to the directory holding the training data
        - label_path : The path to the directory holding the label data
        '''
        
        # Define the class variables to get the 
        # input and label path
        self.input_path = input_path
        self.label_path = label_path

        # Check if the path names ends in '/'
        if self.input_path[-1] != '/':
            self.input_path += '/'
        if self.label_path[-1] != '/':
            self.label_path += '/'

        self.inpn = np.array(sorted(os.listdir(input_path)))
        self.labn = np.array(sorted(os.listdir(label_path)))

        self.tinp = len(self.inpn)
        self.tlab = len(self.labn)
        
    def show_paths(self):
        '''
        The method to show the input path

        Arguments:
        None

        Returns:
        - str = The input path
        - str = The label path
        '''

        return self.input_path, self.label_path

    def glowhigh(self, batch_size=1, return_range=True):
        if batch_size > self.total_inputs:
            raise RuntimeError('Batch size is greater than the total number of files present!!')

        oidx = np.random.randint(0, self.total_inputs, size=1)
        sidx = oidx + batch_size

        if sidx >= self.total_inputs:
            sidx = oidx
            oidx = sidx - batch_size
        
        if return_range:
            return np.arange(start=oidx, 
                             stop=sidx)

        return oidx, sidx

    def get_batch(self, batch_size):
        pass

    def create_loader(self):
        pass

    def check(self):
        # Make sure the input and label paths exists
        if not os.path.exists(self.input_path) or \
           not os.path.exists(self.label_path):
            raise RuntimeError('The path passed in doesn\'t exists!!')
        # Make sure there are same number of inputs as the number of labels
        assert (self.total_inputs == self.total_labels)



"""
ENet: A Deep Neural Network Architecture forReal-Time Semantic Segmentation

"""

import torch
import torch.nn as nn
import copy
import numpy as np

def mupool(ks:int=2, s:int=2, p:int=0):
    return nn.MaxUnpool2d(kernel_size=ks, stride=s, padding=p)

def mpool(ks:int=2, s:int=2, p:int=0, idxs:bool=False):
    return nn.MaxPool2d(kernel_size=ks, stride=s, padding=p, return_indices=idxs)

def conv(ic:int, oc:int, ks:int=3, s:int=1, p:int=0, d:int=1, norm:bool=True, act:nn.Module=nn.PReLU):
    layers = [nn.Conv2d(ic, oc, kernel_size=ks, stride=s, padding=p, dilation=d, bias=not norm)]
    if norm: layers += [nn.BatchNorm2d(oc)]
    if act: layers += [act() if act == nn.PReLU else act(inplace=True)]
    return layers

def uconv(ic:int, oc:int, ks:int=2, s:int=2, p:int=0, op:int=0, norm:bool=True, act:bool=False):
    layers = [nn.ConvTranspose2d(ic, oc, kernel_size=ks, stride=s, padding=p, output_padding=op, \
                                 bias=not norm)]
    if norm: layers += [nn.BatchNorm2d(oc)]
    if act: layers += [nn.ReLU(inplace=True)]
    return layers

class InitialBlock(nn.Module):
    """
    This module is the initial block of the ENet architecture
    """

    def __init__(self, ic=3, oc=16):
        super().__init__()
        self.main = nn.Sequential(*conv(ic, oc-3, 3, 2, 1, norm=False, act=None))
        self.side = mpool()
        self.norm = nn.BatchNorm2d(oc)
        self.prelu = nn.PReLU()
    
    def forward(self, x):
        return self.prelu(self.norm(torch.cat((self.main(x), self.side(x)), dim=1)))

class RDANeck(nn.Module):
    """
    This module implements the Regular, Dilated and Asymmetric Blocks

    Args:
        ic: # of in_channels
        oc: # of out_channels
        d:  Dilation rate. 
            Default - 1
        aflag: Flag to indicate whether it is an asymetric block or not.
               Default - False
        pratio: Ratio by which the channels are reduced and increased in the main branch
                Default - 4
        p: The dropout probability. 
           Default - 0.1
    """

    def __init__(self, ic:int, oc:int, d:int=1, aflag:bool=False, pratio:int=4, 
                 p:float=0.1, act:nn.Module=nn.ReLU, device:str='cuda'):

        super().__init__()
        
        self.cpad = np.abs(oc - ic)
        ks = (1, 5)
        pad = (0, 2)
        rd = ic // pratio
        self.dev = torch.device('cuda' if torch.cuda.is_available() and device=='cuda' \
                                else 'cpu')
        
        l = [*conv(rd, rd, ks, 1, pad, norm=None, act=None), \
             *conv(rd, rd, ks[::-1], 1, pad[::-1])] if aflag \
            else [*conv(rd, rd, 3, 1, d, d)]
        
        # TODO Check if the position of dropout is correct
        self.main = nn.Sequential(*conv(ic, rd, 1, 1, 0), *l, *conv(rd, oc, 1, 1, 0, act=None),
                                  nn.Dropout2d(p))
        self.act = act(inplace=True)

    def forward(self, x):
        ix = x
        x = self.main(x)
        zshape = list(x.shape); zshape[1] = self.cpad
        # TODO Add device for torch.zeros
        out = torch.cat((ix, torch.zeros(*zshape, device=self.dev)), dim=1) + x if self.cpad else ix + x
        return self.act(out)

class DNeck(nn.Module):
    """
    This module implements the Downsampling Block

    Args:
        ic: # of in_channels
        oc: # of out_channels
        pratio: Ratio by which the channels are reduced and increased in the main branch
                Default - 4
        p: The dropout probability. 
           Default - 0.1
    """

    def __init__(self, ic, oc, p=0.1, pratio=4, act:nn.Module=nn.PReLU, device='cuda'):
        super().__init__()

        self.cpad = np.abs(oc - ic)
        rd = ic // pratio
        self.dev = torch.device('cuda' if torch.cuda.is_available() and device=='cuda' \
                                else 'cpu')

        self.main = nn.Sequential(*conv(ic, rd, 1, 1, 0), *conv(rd, rd, 3, 2, 1), 
                                   *conv(rd, oc, 1, 1, 0, act=None))
        self.mpool = mpool(idxs=True)
        self.act = act(inplace=True) if act == nn.ReLU else act()

    def forward(self, x):
        ix, idxs = self.mpool(x)
        x = self.main(x)
        zshape = list(x.shape); zshape[1] = self.cpad
        # TODO Add device for torch.zeros
        out = torch.cat((ix, torch.zeros(*zshape, device=self.dev)), dim=1) + x if self.cpad else ix + x 
        return self.act(out), idxs

class UNeck(nn.Module):
    """
    This module implements the Downsampling Block
    Args:
        ic: # of in_channels
        oc: # of out_channels
        pratio: Ratio by which the channels are reduced and increased in the main branch
                Default - 4
        p: The dropout probability. 
           Default - 0.1
    """

    def __init__(self, ic:int, oc:int, p:float=0.1, pratio:int=4, act:nn.Module=nn.ReLU):
        super().__init__()

        self.cpad = oc - ic
        rd = ic // pratio
        # TODO Check if the position of dropout is correct
        self.main = nn.Sequential(*conv(ic, rd, 1, 1, 0), *uconv(rd, rd, 3, 2, 1, 1),
                                   *conv(rd, oc, 1, 1, 0, act=None), nn.Dropout2d(p))
        
        self.conv = nn.Sequential(*conv(ic, oc, 1, 1, 0))
        self.mupool = mupool()
        self.act = act(inplace=True)

    def forward(self, x, idxs):
        ix = self.mupool(self.conv(x), idxs)
        x = self.main(x)
        return self.act(x + ix)

class ENet(nn.Module):
    """
    This is the implementation of the ENet architecture.
    Reproduced from: https://arxiv.org/pdf/1606.02147.pdf

    Args:
        nc: # of classes
    """

    def __init__(self, nc):
        super().__init__()

        self.nc = nc
        
        self.init = InitialBlock()

        self.d1 = DNeck(16, 64, 0.01)
        self.b1 = nn.Sequential(*[RDANeck(64, 64, p=0.01) for _ in range(4)])
        
        self.d2 = DNeck(64, 128)
        l = [RDANeck(128, 128), RDANeck(128, 128, d=2), RDANeck(128, 128, aflag=True), 
             RDANeck(128, 128, d=4), RDANeck(128, 128), RDANeck(128, 128, d=8), 
             RDANeck(128, 128, aflag=True), RDANeck(128, 128, d=16)]
        self.b2_3 = nn.Sequential(*l, *copy.deepcopy(l))

        self.u1 = UNeck(128, 64)
        self.b4 = nn.Sequential(*[RDANeck(64, 64, act=nn.ReLU) for _ in range(2)])

        self.u2 = UNeck(64, 16)
        self.b5 = nn.Sequential(RDANeck(16, 16, act=nn.ReLU), *uconv(16, nc, 3, 2, 1, 1))

    def forward(self, x):
        o1, idx1 = self.d1(self.init(x))
        o2, idx2 = self.d2(self.b1(o1))
        return self.b5(self.u2(self.b4(self.u1(self.b2_3(o2), idx2)), idx1))



import torch
import numpy as np

import os
import requests
import matplotlib.pyplot as plt

from PIL import Image
from torchvision import transforms as T
from matplotlib.gridspec import GridSpec
from scratchai.nets.style_transfer.image_transformation_net import ITN_ST
from scratchai.datasets.labels import *
from scratchai.pretrained import urls
from scratchai import *


__all__ = ['classify', 'stransfer', 'attack']


def classify(path:str, nstr='resnet18', trf:str=None, gray:bool=False):
  """
  One call to classify an image.

  Arguments
  ---------
  img : str
        The path to the image file / url to the image file.
  nstr : str, nn.Module
         The name of the net to be used in string format.
         or the net itself
  trf : str
        The transforms to be used in the image. 
        The str is passed to utils.get_trf(). 
        Defaults to 'rz256_cc224_tt_normimgnet'
  gray : bool
       Set this to True if the model expects 2d images. Defaults to False.
  Returns
  -------
  pred : str
         The predicted value.

  Notes
  -----
  Given path to an image this function loads it and converts it to a tensor and
  normalizes it using imagenet specific normalization values and passes it to 
  a resnet output the predicted value.
  """
  assert type(trf) == str or type(trf) == type(None)
  if type(nstr) is str and nstr.endswith('_mnist'): gray = True
  
  if not trf:
    if not gray: trf = imgutils.get_trf('rz256_cc224_tt_normimgnet')
    else: trf = imgutils.get_trf('rz32_cc28_tt_normmnist')
  else: trf = imgutils.get_trf(trf)
  
  # Getting the net
  if type(nstr) is str: net = getattr(nets, nstr)()
  else: net = nstr
  net.eval()

  # Getting the image from `path`
  if type(path) == str:
    # Special Case: if net == 'lenet_mnist' then image needs to have one channel
    if gray: img = trf(imgutils.gray(path)).unsqueeze(0)
    # Normal Cases
    else: img = trf(imgutils.load_img(path)).unsqueeze(0)
  else:
    img = trf(path).unsqueeze(0)

  # Edge Case: In case of 2D Images, the above line adds the channel dim
  # And then the batch dim needs to be added.
  if len(img.shape) == 3: img = img.unsqueeze(0)
  
  # Getting the image from net
  val, pred = torch.max(net(img), dim=1)
  val = val.item(); pred = pred.item()
  
  # In case the net is trained on mnist
  if gray: label = mnist_labels[pred]
  # In case the net is trained on Imagenet
  else: label = imagenet_labels[pred]
  
  return label, val


def stransfer(path:str, style:str=None, save:bool=False, show:bool=True):
  """
  One call to transfer the style of an image.

  Arguments
  ---------
  img : str
        The path to the image file / url to the image file.
  style : str
          The style in which to convert. Defaults to None.
          If None, it picks up a style in random.
  save : bool
         If True, saves the image to disk. Default False.
  show : bool
         If true, it shows the image.
  Returns
  -------
  pred : str
         The predicted value.

  Notes
  -----
  Given path to an image this function loads it and converts it to a tensor and
  normalizes it using imagenet specific normalization values and passes it to 
  a resnet output the predicted value.
  """
  avbl_styles = ['elephant_skin', 'snake_skin']
  if style is None: style = np.random.choice(avbl_styles, 1)[0]
  sdict = utils.load_from_pth(getattr(urls, style + '_url'), style)

  trf = T.Compose([T.ToTensor(),
                   T.Lambda(lambda x : x.mul(255))])

  if type(path) == str:
    img = trf(imgutils.load_img(path))
  else:
    img = trf(path)

  net = ITN_ST(); net.load_state_dict(sdict); net.eval()
  out = net(img.unsqueeze(0)).squeeze().cpu()
  out = Image.fromarray(out.squeeze().transpose(0, 1).transpose(1, 2).detach()
             .clone().cpu().clamp(0, 255).numpy().astype('uint8'))
  if save: imgutils.imsave(out)
  if show: imgutils.imshow(out)
  return out


def attack(x, atk=attacks.FGM, nstr='resnet18', ret:bool=False, **kwargs):
  """
  One call to perform an attack on an image.

  Arguments
  ---------
  x : str, torch.Tensor
      The input image.
  atk : scratchai.attack.attack
        The attack which to perform. Defaults to FGM attack.
  nstr : str
         The net to use. (if needed) Defaults to None.
        For black box attacks, no need of passing the net.
  ret : bool
        If true, it returns the original image, adversarial image and others
        like the true label, adversarial label and so on.
  """
  # Checks
  assert not atk == attacks.Noise, 'Noise attack not supported due to a diff '\
                                   'preprocessing pipeline for noise attacks.'\
                                   ' Will support in future'
  
  if 'y' in kwargs:
    if isinstance(kwargs['y'], int): 
      kwargs['y'] = torch.Tensor([kwargs['y']]).long()
    assert kwargs['y'].item() >= 0  and kwargs['y'].item() < 1000, 'The class'\
            'label must be between [0, 1000) where the number represents the'\
            'class label from the available classes in imagenet.'

  x = imgutils.load_img(x) if isinstance(x, str) else x
  tlabl = classify(x, nstr) if nstr is not None else classify(x)
  net = getattr(nets, nstr)().eval(); utils.freeze(net)
  atk = atk(net=net, **kwargs)

  # Preprocess image for attack
  trf = imgutils.get_trf('rz256_cc224_tt_normimgnet')
  advx = atk(trf(x).unsqueeze(0))

    
  ## TODO Uncomment these 2 lines and comment the 2 lines below these
  ## after figuring out how to convert the adversarial image to PIL 
  ## and retain the precision so the PIL image when converted back to tensor
  ## is misclassified correctly as the orginial adversarial image.
  #advx_pil = T.ToPILImage()(advx.squeeze())
  #plabl = classify(advx_pil, nstr) if nstr is not None else classify(advx_pil)
  val, pred = torch.max(net(advx), dim=1)
  plabl = (imagenet_labels[pred.item()], val.item())

  limgs = [trf(x), None, advx]
  limgs[1] = imgutils.diff_imgs(limgs[0], limgs[2])
  limgs[0] = imgutils.unnorm(limgs[0]); limgs[2] = imgutils.unnorm(limgs[2])
  ltils = ['Original Image\nPred: {} \nwith {:.1f} confidence.'
              .format(tlabl[0].split(',')[0], tlabl[1]), 'Difference', 
           'Adversarial Image\nPred: {} \nwith {:.1f} confidence.'
              .format(plabl[0].split(',')[0], plabl[1])]
  
  if ret: return limgs, tlabl, plabl

  # TODO Just go over the how the images are viz ones
  # in case something is weird.
  plt.figure(figsize=(10, 10))
  gs = GridSpec(1, 3)
  gs.update(wspace=0.025, hspace=0.0)
  
  for ii in range(3):
    plt.subplot(gs[ii])
    plt.axis('off')
    plt.title(ltils[ii])
    plt.imshow(imgutils.t2i(limgs[ii]))
  plt.show()



import torch
import torch.nn as nn
import unittest

from torchvision import transforms as T
from scratchai.datasets.labels import imagenet_labels
from scratchai import *

class TestOneCalls(unittest.TestCase):
  
  url_1 = 'https://cdn.instructables.com/FVI/FJPH/FWYHRVGC/FVIFJPHFWYHRVGC.'\
          'LARGE.jpg'
  lab_1 = 'rain barrel'

  url_2 = 'http://bradleymitchell.me/wp-content/uploads/2014/06/decompressed.'\
          'jpg'
  lab_2 = 'five'

  def test_classify(self):
    """
    This function ensures the classify function is working properly.
    """
    ns = ['alexnet', 'resnet18', 'resnet34', 'resnet50', 'resnet101', 
         'resnet152']
    
    # Check that url works
    for n in ns:
      pred, val = one_call.classify(TestOneCalls.url_1, nstr=n)
      self.assertTrue(pred == TestOneCalls.lab_1, 'Doesn\'t Work!')

    pred, val = one_call.classify(TestOneCalls.url_1)
    self.assertTrue(isinstance(pred, str), 'Doesn\'t Work!')
    self.assertTrue(pred == TestOneCalls.lab_1, 'Doesn\'t Work!')
    # TODO Check that path in local works
    
    # Checks for one_classify can take nn.Module as argument
    net = nets.resnet18()
    pred, val = one_call.classify(TestOneCalls.url_1, nstr=net)
    self.assertTrue(isinstance(pred, str), 'Doesn\'t Work!')
    self.assertTrue(pred == TestOneCalls.lab_1, 'Doesn\'t Work!')

    # Check that mnist works
    mnist_ns = ['lenet_mnist', 'alexnet_mnist']
    for n in mnist_ns:
      pred, val = one_call.classify(TestOneCalls.url_2, nstr=n, 
                                    trf='rz32_cc28_tt')
      pred, val = one_call.classify(TestOneCalls.url_2, nstr=n)

      self.assertTrue(isinstance(pred, str), 'Doesn\'t Work!')
      self.assertTrue(pred == TestOneCalls.lab_2, 'Doesn\'t Work!')

    self.assertRaises(AssertionError, lambda: one_call.classify(
            TestOneCalls.url_1, trf=imgutils.get_trf('tt_normmnist')))
    
    # Checks for one_classify and MNIST that it can take nn.Module as input
    # Given that gray is set to True for 2D images.
    net = nets.lenet_mnist()
    pred, val = one_call.classify(TestOneCalls.url_2, nstr=net, gray=1)
    self.assertTrue(isinstance(pred, str), 'Doesn\'t Work!')
    self.assertTrue(pred == TestOneCalls.lab_2, 'Doesn\'t Work!')

  def test_stransfer(self):
    """
    Ensures the stransfer function is working properly.
    """
    tt = T.ToTensor()
    imgshape = tt(imgutils.load_img(TestOneCalls.url_1)).shape
    simgshape = tt(one_call.stransfer(TestOneCalls.url_1, show=False)).shape
    self.assertTrue(imgshape == simgshape, 'doesn\'t look good')

  def test_attack(self):
    """
    Ensures the one_call.attack function is working as expected.
    """
    atks = [attacks.FGM, attacks.PGD, attacks.Semantic]
    for atk in atks:
      limgs, tlabl, plabl = one_call.attack(TestOneCalls.url_1, atk=atk, ret=1)
      self.assertFalse(tlabl == plabl, 'attack didn\t work')
      self.assertTrue(limgs[1].sum() != 0., 'attack didn\t work')
    # Noise attack not supported through one_call
    self.assertRaises(AssertionError, lambda: \
        one_call.attack(TestOneCalls.url_1, atk=attacks.Noise))
    # Label should be between 0 and 1000 (1000 excluded)
    self.assertRaises(AssertionError, lambda: \
        one_call.attack(TestOneCalls.url_1, atk=attacks.PGD, y=1000))

    
    y = torch.randint(0, 1000, ())
    limgs, tlabl, plabl = one_call.attack(TestOneCalls.url_1, atk=attacks.PGD, 
                                   y=y.item(), ret=True)
    self.assertTrue(plabl[0] == imagenet_labels[y.item()], 'not working!')



import torch
import torch.nn as nn
import numpy as np
import os

import cv2
from torchvision import transforms
from subprocess import call
from scratchai._config import home


__all__ = ['load_from_pth', 'implemented', 'name_from_object', 'setatrib',
           'load_pretrained', 'Topk', 'freeze', 'AvgMeter', 'count_params',
           'gpfactor', 'sgdivisor']


def count_params(net):
  return sum(p.numel() for p in net.parameters() if p.requires_grad)


# No Tests written for this function.
def cam():
  c = cv2.VideoCapture(0)
  while True:
    r, img = c.read()
    cv2.imshow('cam', img)
    if cv2.waitKey(1) == 113: break
  cv2.destroyAllWindows()


def load_from_pth(url, fname='random', key='state_dict'):
  """
  Function to download/load the pth file and return the key mentioned.

  Arguments
  ---------
  url : str
        The url from which to download the file.
  key : str
        The key of the dict to return
  fname : str
          The name with which the file should be saved.
  Returns
  -------
  val : the type of element stored in the key (mostly torch.tensor)
        The value.
  """
  
  # TODO Download this file to a location where it doesn't need to 
  # be downloaded again and again.
  prefix = home
  if not os.path.isfile(prefix + fname + '.pth'):
    call(['wget', '-O', '{}{}.pth'.format(prefix, fname), url])
  ckpt = torch.load('{}{}.pth'.format(prefix, fname), map_location='cpu')
  return ckpt[key] if key in ckpt else ckpt


def implemented(module, func):
  """
  Function to check if a function func exists in module.

  Arguments
  ---------
  module : any
           The module where the presence of function needs
           to be checked.
  func : str
         The name of the function in str whose presence 
         needs to be confirmed.

  Raises
  ------
  NotImplementedError : If the function is not present in module.
  """
  imp = getattr(module, func)
  if not callable(imp):
    raise NotImplementedError


def name_from_object(obj):
  """
  This function returns the name of the object
  from its initialized instance.

  Arguments
  ---------
  obj : any
        The object whose name needs to be returned.

  Returns
  -------
  name : str
         The name of the object
  """
  # str(obj,__class__) -> "<class 'torch.optim.adam.Adam'>"
  cls_str = str(obj.__class__) if str(obj.__class__) != str(type) else str(obj)
  # [1:-2] -> "class 'torch.optim.adam.Adam"
  # .split('.')[-1] -> "Adam"
  return cls_str[1:-2].split('.')[-1].lower()


def setatrib(obj, attr:str, val):
  """
  Overiding the default getattr

  Arguments
  ---------
  obj : object
        The object from which to get the attribute.
  attr : str
         The attribute in string format.
  """
  attrs = attr.split('[')
  for ii in range(len(attrs)):
    # ie its a list
    if attrs[ii][-1] == ']': attr = attrs[ii][:-1]
    else: attr = attrs[ii]

    if ii == len(attrs)-1:
      setattr(obj, attr, val)
    else:
      obj = getattr(obj, attr)


def load_pretrained(net:nn.Module, url:str, fname:str, nc:int=None, attr='fc',
                    inn:int=512):
  """
  Helps in Loading Pretrained networks
  """
  net.load_state_dict(load_from_pth(url, fname))
  # If nc != None, load a custom last linear layer
  # After freezing the rest of the network
  if nc is not None:
    for p in net.parameters():
      p.requires_grad_(False)
    setatrib(net, attr, nn.Linear(inn, nc))
  return net


def freeze(net:nn.Module):
  """
  Freeze the net.

  Arguments
  ---------
  net : nn.Module
        The net to freeze

  """
  for p in net.parameters():
    if p.requires_grad:
      p.requires_grad_(False)


class AvgMeter():
  """
  Computes and stores the current avg value.

  Arguments
  ---------
  name : str
         The name of the meter
  fmt : str
        The format in which to show the results

  Notes
  -----
  When you call an instance of this class, make sure to call it
  with (val/cnt, cnt) where the val is already divided by cnt.
  """
  def __init__(self, name, fmt=':.2f'):
    self.name = name
    self.fmt = fmt
    self.reset()
  
  def __call__(self, val, cnt):
    self.val = val
    self.sum += val * cnt
    self.cnt += cnt
    self.avg = self.sum / self.cnt

  def reset(self):
    self.val = 0.; self.sum = 0.
    self.cnt = 0.; self.avg = 0.
  
  def __str__(self):
    return '{name} - {avg}'.format(**self.__dict__)


class Topk():
  """
  A class that helps easily maintain the topk values.
  And maintains a seperate AvgMeter for each k.

  Arguments
  ---------
  topk : tuple
         The topk values that needs to be handled.
  """
  def __init__(self, name:str, topk:tuple=(1,)):
    assert 0 not in topk
    self.name = name
    self.topk = tuple(sorted(set(topk)))
    self.ks = len(self.topk)
    self.avgmtrs = {}
    for k in topk:
      n = name + str(k)
      self.avgmtrs[n] = AvgMeter(n)

  def update(self, vals, cnt):
    """
    Updates all the meters and values.
    
    Arguments
    ---------
    vals : tuple, list
           Stores all the values for each k. Even if there's 
           just 1 k, pass it as a tuple. Assumes all the passed
           value are in sorted fashion, like the first element is
           for first k, the second element for the second k and so on.
    cnt : int, float
          The total number of elements.
    """
    assert len(vals) == self.ks
    for i, val in enumerate(vals):
      self.avgmtrs[self.name + str(int(self.topk[i]))](val, cnt)
      
  def __str__(self):
    s = ''
    for name, mtr in self.avgmtrs.items():
      s += 'Top {} {} is {}\n'.format(name[-1], self.name, mtr.avg)
    return s


def gpfactor(num:int):
  """
  Function that returns the greatest prime factor.

  Arguments
  ---------
  num : int
        The integer whose greatest prime factor is needed.

  Returns
  -------
  num : int
        The greatest prime factor.
  """
  # TODO Implement this function in cpp and wrap it up with python
  assert isinstance(num, int) == True
  assert num != 0
  mf = -1
  while num % 2 == 0: mf = 2; num >>= 1
  for i in range(3, int(np.sqrt(num))+1, 2):
    while num % i == 0: mf = i; num //= i
  if num > 2: mf = num
  return mf
  

def sgdivisor(num:int):
  """
  Function that returns the smallest and the largest divisor of a number.

  Arguments
  ---------
  num : int
        The integer whose divisors are needed.

  Returns
  -------
  num : 2-tuple of integers
        The smallest and the largest divisors of the number.
  """
  # TODO Implement this function in cpp and wrap it up with python
  assert isinstance(num, int) == True
  if num == 0: return (0, 0)
  sdiv = 1
  if num % 2 == 0: sdiv = 2
  else:
    for ii in range(3, int(np.sqrt(num))+1, 2):
      if num % ii == 0:
        sdiv = ii
        break
  mdiv = num // sdiv
  return sdiv, mdiv


def count_modules(net:nn.Module):
  """
  TODO
  """
  allm = []
  mdict = {}
  for m in net.modules():
    name = m.__class__.__name__
    if name in allm:
      mdict[name] += 1
    else:
      allm.append(name)
      mdict[name] = 1

  return mdict



import torch
import numpy as np
import matplotlib.pyplot as plt
import cv2
import operator
import PIL
import os
import requests

from mpl_toolkits.mplot3d import Axes3D
from PIL import Image
from torchvision import transforms as T

from scratchai import utils


__all__ = ['thresh_img', 'mask_reg', 'mark_pnt_on_img', 'load_img', 't2i', 
           'imsave', 'imshow', 'unnorm', 'get_trf', 'surface_plot', 'gray',
           'diff_imgs', 'mean', 'std']


def thresh_img(img:np.ndarray, rgb, tcol:list=[0, 0, 0]):
  """
  This function is used to threshold the image to a certain color.

  Args:
    img (np.ndarray): The image
    rgb (tuple or list): The color below which all colors will be blacked.
    tcol (list): The color to be filled in the thresholded part of
          the image. Defaults to [0, 0, 0].

  Returns:
    np.ndarray: The threshold image
  """

  assert type(img) == np.ndarray
  assert type(rgb) == tuple or type(rgb) == list
  assert type(tcol) == list

  img = np.copy(img)
  tidx = (img[:, :, 0] < rgb[0]) \
     | (img[:, :, 1] < rgb[1]) \
     | (img[:, :, 2] < rgb[2])
  img[tidx] = tcol
  return img


def mask_reg(img, pnts, reln, deg:int=1, tcol:tuple=(0, 0, 0), 
      locate:bool=False, invert:bool=False) -> np.ndarray:
  """
  Region Masking.

  Given a set of points which corrsponds to a polygon this
  function masks that polygon and returns the masked image.
  
  Arguments
  ---------
  img : np.ndarray, shape (H, W, 3)
     The image which needs to be masked.
  pnts : list
     A list containing the set of points (x, y)
     where each is a list / tuple.
  reln : list
     Where each element is a list/tuple (n1, n2)
     where n1 < len(points) and n2 < len(points) and it denotes
     that these two points are connected. 
     Do note: that n1 corresponds to the n1th element in points.
  deg : int
     Degree of the fitting polynomial.
  tcol : tuple of 3 elements
     A list containing the rgb values for the
     color which the mask needs to be filled with.
     Defaults to (0, 0, 0)
  locate : bool
      Denotes if to mark the points also in the image. 
      Defaults to False.
  invert : bool
      Denotes whether to invert the final indexes. Defaults to False.

  Returns
  -------
  img : np.ndarray
     The masked image.
  """

  assert type(img) == np.ndarray
  for param in [pnts, reln, tcol]:
    assert type(param) == list or type(param) == tuple
  
  img = np.copy(img)
  h, w = img.shape[1], img.shape[0]
  ops = {'<' : operator.lt, '>' : operator.gt}
  
  # Fit the lines
  psolns = {}
  for ii, (p1, p2, rel) in enumerate(reln):
    pnt1 = pnts[p1]; pnt2 = pnts[p2]
    soln = list(np.polyfit((pnt1[0], pnt2[0]), (pnt1[1], pnt2[1]), deg))
    soln.append(rel)
    psolns[ii] = soln

  # Find the region inside the lines
  xx, yy = np.meshgrid(np.arange(h), np.arange(w))
  for ii, val in psolns.items():
    # TODO Give option to choose btw bitwise_and and bitwise_or
    reg_thresh = ops[val[2]](yy, (xx*val[0] + val[1])) if not ii else \
          reg_thresh & ops[val[2]](yy, (xx*val[0] + val[1]))
      
  # Color pixels which are inside the region of interest
  img[reg_thresh == False if invert else reg_thresh] = tcol
  
  return img if not locate else mark_pnt_on_img(img, pnts)


def mark_pnt_on_img(img, pnts:list, col:tuple=(0, 255, 0)) -> np.ndarray:
  """
  Mark points on Image.

  Arguments
  ---------
  img : np.ndarray - Shape like (H x W x 3)
     The image on which the points need to be marked.
  pnts : list of tuples
     where each tuple is (x, y) a point to mark.
  col : tuple of 3 elements
     which are the rgb values for the color of the point.
  Returns
  -------
  img : np.ndarray
     With all the points marked on the image
  """

  img = np.copy(img)
  for pnt in pnts:
    cv2.circle(img, pnt, 10, col, -1)
  return img


def load_img(path:str, rtype=PIL.Image.Image):
  """
  Helper function to load a image from url as well as path.

  Arguments
  ---------
  path : str
        The file path or url from where to load the image.
  
  rtype : data type object
          The type of object in which to return the img.

  Returns
  -------
  img : rtype
        where rtype is the type of object in which to return the object.
  """
  # Other types not supported
  assert rtype in [PIL.Image.Image, np.ndarray]

  # Check if url
  # TODO Improve check
  is_url = path.startswith('http')
  fname = path.split('/')[-1]
  if is_url:
    if not os.path.isfile('/tmp/{}'.format(fname)):
      with open('/tmp/'+fname, 'wb') as f:
        f.write(requests.get(path).content)
    path = '/tmp/'+fname
  
  img = Image.open(path).convert('RGB')
  
  if rtype is np.ndarray:
    return np.array(img)
  return img


def t2i(img, rt=PIL.Image.Image, no255=False):
  """
  Converts torch.Tensor images to PIL images.

  Arguments
  ---------
  img : torch.Tensor
        The tensor image which to convert.
  rt : type
       The type of Image to be returned
  Returns
  -------
  img : PIL.Image.Image
        The converted PIL Image

  Notes
  -----
  Expects a [1 x 3 x H x W] torch.Tensor or [3 x H x W] torch.Tensor
  Converts it to a PIL.Image.Image of [H x W x 3]

  Converting to PIL.Image.Image losses precision if source image
  is of type float
  """
  out= img.squeeze().transpose(0, 1).transpose(1, 2).detach().clone().cpu()
  if not no255: out = out.mul(255).clamp(0, 255)
  out = out.numpy()
  if rt == PIL.Image.Image:
    # Note: .astype('uint8') losses precision
    return Image.fromarray(out.astype('uint8'))
  elif rt == np.ndarray:
    return out


def imsave(img, fname='random.png'):
  """
  Helper function to save an image to disk.

  Arguments
  ---------
  img : PIL.Image.Image, torch.Tensor
        The image to save.
  
  fname : str
          File Name. Defaults to random.png
  """

  if isinstance(img, torch.Tensor): img = t2i(img)
  img.save(fname)


def imshow(img, normd:bool=False, rz=224, **kwargs):
  """
  Display image.

  Arguments
  ---------
  img : torch.Tensor, PIL.Image.Image, list
        The image to display
  normd : bool
          If True, and if img is torch.Tensor then it unnormalizes the image
          Defaults to False.
  rz : int, tuple
       Row and Cols to resize to. If int, its taken as (rz, rz)
       Defaults to 224.
  """
  if isinstance(rz, int): rz = (rz, rz)
  if isinstance(img, list):
    nimgs = len(img)
    fig = plt.figure(figsize=(8, 8))
    # TODO Update to handle when nimgs is a prime and a few other cases
    gp = utils.gpfactor(nimgs)
    odiv = nimgs // gp
    col, row = gp, odiv

    for i in range(1, row*col+1):
      fig.add_subplot(row, col, i)

      cimg = img[i-1]
      if isinstance(cimg, torch.Tensor):
        cimg = t2i(unnorm(cimg) if normd else cimg, **kwargs)
      if rz is not None: cimg = cimg.resize(rz, Image.ANTIALIAS)

      plt.axis('off'); plt.imshow(cimg)

  elif isinstance(img, torch.Tensor):
    img = t2i(unnorm(img) if normd else img, **kwargs)
    plt.imshow(img)

  else: plt.imshow(img)
  plt.show()


def unnorm(t:torch.Tensor, mean=None, std=None):
  """
  Given an image this unnormalizes the image and returns it

  Arguments
  ---------
  t : torch.Tensor
      The image to unnormalize
  mean : list,(in case of >1 channels) else float
         The mean to multiply to the image
         Defaults to: [0.485, 0.456, 0.406],
  std : list,(in case of >1 channels) else float
         The std to add to the image
         Defaults to: [0.229, 0.224, 0.225]

  Returns
  -------
  t : torch.Tensor
      The unnormalized image.
  """
  if mean is None: mean = torch.Tensor([0.485, 0.456, 0.406])
  if std is None: std = torch.Tensor([0.229, 0.224, 0.225])
  # Not sure how to change the dim while multiplying so
  # changing the channel dimension from 0 to 2
  # Performing the operations and then changing it back
  t = (t.squeeze().transpose(0, 1).transpose(1, 2) * std) + mean
  return t.transpose(2, 1).transpose(1, 0)


def mean(t):
  """
  Calculates and returns the mean of a tensor

  Arguments
  ---------
  t : torch.Tensor
      The tensor whose mean is to be calculated
  
  Returns
  -------
  m : float
      The mean of the tensor
  """
  if isinstance(t, torch.Tensor): t = t.numpy()
  return t.sum() / t.size


def std(t):
  """
  Calculates and returns the standard deviation of a tensor

  Arguments
  ---------
  t : torch.Tensor
      The tensor whose std is to be calculated
  
  Returns
  -------
  m : float
      The std of the tensor
  """
  if isinstance(t, torch.Tensor): t = t.numpy()
  return np.sqrt(((t - mean(t)) ** 2).sum() / t.size)


def gray(img):
  """
  Converts RGB Image to Grayscale.
  
  Arguments
  ---------
  img : str, PIL.Image.Image
        The Image which is to be grayscaled.

  Returns
  -------
  img : PIL.Image.Image
        The grayscaled Image.

  References
  ----------
  https://stackoverflow.com/a/12201744
  """

  if isinstance(img, str):
   img = load_img(img)
  img = np.array(img)
  return Image.fromarray(np.dot(img[..., :3], [0.2989, 0.5870, 0.1140]))


def diff_imgs(img1, img2, show=False, **kwargs):
  """
  Returns the difference of 2 images.

  Arguments
  ---------
  img1 : torch.Tensor, [C x H x W]
         The image from which to be subtracted.
  img2 : torch.Tensor, [C x H x W]
         The image which to be substracted.
  """
  dimg = img1.float().squeeze() - img2.float().squeeze()
  if show: imshow(dimg, **kwargs)
  return dimg


def surface_plot(matrix:np.ndarray):
  """
  Function to make a surface plot for a 2D matirx.

  Arguemnts
  ---------
  matrix : np.array
           The matrix for which the surface needs to be plot.
  """
  x, y = np.meshgrid(np.arange(matrix.shape[0]), np.arange(matrix.shape[1]))
  fig = plt.figure()
  ax = fig.add_subplot(111, projection='3d')
  surf = ax.plot_surface(x, y, matrix, cmap=plt.cm.coolwarm)
  #fig.colorbar(surf)
  plt.show()


def get_trf(trfs:str):
  """
  A function to quickly get required transforms.

  Arguments
  ---------
  trfs : str
         An str that represents what T are needed. See Notes

  Returns
  -------
  trf : torch.transforms
        The transforms as a transforms object from torchvision.

  Notes
  -----
  >>> get_trf('rz256_cc224_tt_normimgnet')
  >>> T.Compose([T.Resize(256),
                          T.CenterCrop(224),
                          T.ToTensor(),
                          T.Normalize([0.485, 0.456, 0.406], 
                                      [0.229, 0.224, 0.225])])
  """
  # TODO Write tests
  # TODO Add more options
  trf_list = []
  for trf in trfs.split('_'):
    if trf.startswith('rz'):
      val = (int(trf[2:]), int(trf[2:]))
      trf_list.append(T.Resize(val))
    elif trf.startswith('cc'):
      val = (int(trf[2:]), int(trf[2:]))
      trf_list.append(T.CenterCrop(val))
    elif trf.startswith('rr'):
      trf_list.append(T.RandomRotation(int(trf[2:])))
    elif trf.startswith('rc'):
      trf_list.append(T.RandomCrop(int(trf[2:])))
    # TODO Add other padding modes
    elif trf.startswith('pad'):
      trf_list.append(T.Pad(int(trf[3:]), padding_mode='reflect'))
    elif trf.startswith('rhf'):
      val = float(trf[3:]) if trf[3:].strip() != '' else 0.5
      trf_list.append(T.RandomHorizontalFlip(val))
    elif trf.startswith('rvf'):
      val = float(trf[3:]) if trf[3:].strip() != '' else 0.5
      trf_list.append(T.RandomVerticalFlip(val))
    # T.ColorJitter
    # TODO Add a way to specify all three values
    # As of we take just 1 value and pass 3 equal ones.
    elif trf.startswith('cj'):
      val = [float(trf[2:])] * 3
      trf_list.append(T.ColorJitter(*val))
    elif trf == 'tt':
      trf_list.append(T.ToTensor())
    elif trf == 'normimgnet':
      trf_list.append(T.Normalize([0.485, 0.456, 0.406],
                                  [0.229, 0.224, 0.225]))
    elif trf == 'normmnist':
      trf_list.append(T.Normalize((0.1307,), (0.3081,)))
    elif trf == 'fm255':
      trf_list.append(T.Lambda(lambda x : x.mul(255)))
    else:
      raise NotImplementedError

  return T.Compose(trf_list)



"""
The Resnet
"""

import torch
import torch.nn as nn
import scratchai.pretrained.urls as urls

from scratchai.utils import load_from_pth, load_pretrained
from scratchai.nets.blocks.resblock import Resnext


__all__ = ['resnet18_mnist', 'resnet18', 'resnet34', 'resnet50', 'resnet50', 
           'resnet101', 'resnet152', 'resnext18', 'resnext34', 'resnext50', 
           'resnext101', 'resnext152']


def conv(ic:int, oc:int, ks:int=3, s:int=1, p:int=1, norm:nn.Module=nn.BatchNorm2d, act:bool=True):
  layers = [nn.Conv2d(ic, oc, kernel_size=ks, stride=s, padding=p, bias=not norm)]
  if norm: layers += [norm(oc)]
  if act: layers += [nn.ReLU(inplace=True)]
  return layers


class resblock(nn.Module):
  """
  This module is the implementation of a resnet block.

  Arguments
  ---------
  ic : int
       # of input channels
  oc : int
       # of output channels
  ks : int
       kernel_size of the mid_conv
  s : int
      stride of the mid_conv
  p : int
      padding of the mid_conv
  norm : nn.Module
         The Normalization to be used.
  act : bool
        whether to use activation
  pratio : int
           the factor by which the # of in_channels is reduced
  layers : int
           IDEA layers is not yet implemented but the idea is that the
           main branch will contain conv layers equal to the number of layers.
           Helpful in research purposes.
  """

  def __init__(self, ic:int, oc:int=None, norm:nn.Module=nn.BatchNorm2d, 
        act:bool=True, dflag:bool=False, pratio:int=4, layers:int=2, 
        btype:str='basic', fdown:bool=False):
    super().__init__()
    assert btype in ['basic', 'bottleneck']
    self.dflag = dflag
    s = 2 if dflag else 1; oc = ic*2 if not oc else oc
    
    if btype == 'basic':
        self.main = nn.Sequential(*conv(ic, oc, s=s), *conv(oc, oc, act=None))
    elif btype == 'bottleneck':
        interc = oc // 4
        self.main = nn.Sequential(*conv(ic, interc, 1, 1, 0), \
                                  *conv(interc, interc, s=s),
                                  *conv(interc, oc, 1, 1, 0, act=None),)
    self.act = nn.ReLU(inplace=True)
    
    # HACK fdown is introduced just because the 1st layer of a resnet >= 50
    # needs a side branch with a 1x1 convolution. Find a better way if possible.
    if dflag or fdown:
      self.side = nn.Sequential(*conv(ic, oc, 1, s, 0, act=None))
    else:
      self.side = None

  def forward(self, x):
    return self.act(self.main(x) + (self.side(x) if self.side else x))


def res_stage(block:nn.Module, ic:int, oc:int, num_layers:int, dflag:bool=True,
              btype:str='basic', fdown:bool=False):

  """

  Arguments
  ---------
  
  block : nn.Module
          the block type to be stacked one upon another
  ic : int
       # of input channels
  oc : int
       # of output channels
  num_layers - int
               # of blocks to be stacked
  dflag : bool
          Whether the first resblock needs to perform downsampling.
          Defaults to True.
  btype : str, should be one of ['basic'. 'bottleneck']
          The type of resblock to be used. Defaults to 'basic'
  fdown : bool
          If true the side branch *must* contain a conv block,
          whether it performs downsampling or not. Defaults to False.

  Returns
  -------
  layers : list
           A list containing all the nn.Module that is required for this layer.
  """
  
  layers = [block(ic, oc, dflag=dflag, btype=btype, fdown=fdown)]
  layers += [block(oc, oc, btype=btype) for i in range (num_layers-1)]
  return layers


class Resnet(nn.Module):
  """
  The class that defines the ResNet module.

  Arguments:
  nc : int
       # of classes
  ic : int
       # of input channels to the network
  oc1 : int
        # of channels for the output of the first stage
  layers : TOFILL
  lconv : int
          # of conv layers in each Residual Block
  """

  def __init__(self, layers:list, nc:int=1000, lconv:int=2, ex:int=1, 
       block:nn.Module=resblock, ic:int=3, oc1:int=64, **kwargs):
    super().__init__()

    layers = [*conv(ic, oc1, 7, 2, 3),
              nn.MaxPool2d(kernel_size=3, stride=2, padding=1),
              *res_stage(block, oc1, oc1*ex, layers[0], dflag=False, **kwargs),
              *res_stage(block, oc1*ex, oc1*ex*2, layers[1], **kwargs),
              *res_stage(block, oc1*ex*2, oc1*ex*4, layers[2], **kwargs),
              *res_stage(block, oc1*ex*4, oc1*ex*8, layers[3], **kwargs),
              nn.AdaptiveAvgPool2d((1, 1))]
    self.net = nn.Sequential(*layers)
    
    self.fc = nn.Linear(512*ex, nc)

  def forward(self, x):
    bs = x.size(0)
    x = self.net(x)
    x = x.view(bs, -1)
    x = self.fc(x)
    return x


def resnet18_mnist(pretrained=False, **kwargs):
  kwargs['layers'] = [2, 2, 2, 2]
  kwargs['ic'] = 1
  net = Resnet(**kwargs)
  '''
  if pretrained:
    # TODO check inspect module and change the fname
    net.load_state_dict(load_from_pth(urls.resnet18_url, 'resnet18'))
  '''
  return net
  
def resnet18(pretrained=True, **kwargs):
  kwargs['layers'] = [2, 2, 2, 2]
  cust_nc = None
  if pretrained and 'nc' in kwargs: cust_nc = kwargs['nc']; kwargs['nc'] = 1000
  net = Resnet(**kwargs)
  if pretrained:
    # TODO check inspect module and change the fname
    return load_pretrained(net, urls.resnet18_url, 'resnet18', nc=cust_nc)
  return net

def resnet34(pretrained=True, **kwargs):
  kwargs['layers'] = [3, 4, 6, 3]
  cust_nc = None
  if pretrained and 'nc' in kwargs: cust_nc = kwargs['nc']; kwargs['nc'] = 1000
  net = Resnet(**kwargs)
  if pretrained:
    return load_pretrained(net, urls.resnet34_url, 'resnet34', nc=cust_nc)
  return net

def resnet50(pretrained=True, **kwargs):
  kwargs['layers'] = [3, 4, 6, 3]
  kwargs['btype'] = 'bottleneck'
  kwargs['ex'] = 4; kwargs['fdown'] = True
  cust_nc = None
  if pretrained and 'nc' in kwargs: cust_nc = kwargs['nc']; kwargs['nc'] = 1000
  net = Resnet(**kwargs)
  if pretrained:
    return load_pretrained(net, urls.resnet50_url, 'resnet50', nc=cust_nc, 
                           inn=512*kwargs['ex'])
  return net

def resnet101(pretrained=True, **kwargs):
  kwargs['layers'] = [3, 4, 23, 3]
  kwargs['btype'] = 'bottleneck'
  kwargs['ex'] = 4; kwargs['fdown'] = True
  cust_nc = None
  if pretrained and 'nc' in kwargs: cust_nc = kwargs['nc']; kwargs['nc'] = 1000
  net = Resnet(**kwargs)
  if pretrained:
    return load_pretrained(net, urls.resnet101_url, 'resnet101', nc=cust_nc,
                           inn=512*kwargs['ex'])
  return net

def resnet152(pretrained=True, **kwargs):
  kwargs['layers'] = [3, 8, 36, 3]
  kwargs['btype'] = 'bottleneck'
  kwargs['ex'] = 4; kwargs['fdown'] = True
  cust_nc = None
  if pretrained and 'nc' in kwargs: cust_nc = kwargs['nc']; kwargs['nc'] = 1000
  net = Resnet(**kwargs)
  if pretrained:
    return load_pretrained(net, urls.resnet152_url, 'resnet152', nc=cust_nc,
                           inn=512*kwargs['ex'])
  return net

# FIXME The resnet blocks work okay but resnext needs a check
def resnext18(nc, **kwargs):
  kwargs['nc'] = nc
  kwargs['layers'] = [2, 2, 2, 2]
  kwargs['lconv'] = 2
  kwargs['block'] = Resnext
  return Resnet(**kwargs)

def resnext34(nc, **kwargs):
  kwargs['nc'] = nc
  kwargs['layers'] = [3, 4, 6, 3]
  kwargs['lconv'] = 2
  kwargs['block'] = Resnext
  return Resnet(**kwargs)

def resnext50(nc, **kwargs):
  kwargs['nc'] = nc
  kwargs['layers'] = [3, 4, 6, 3]
  kwargs['lconv'] = 3
  kwargs['block'] = Resnext
  return Resnet(**kwargs)

def resnext101(nc, **kwargs):
  kwargs['nc'] = nc
  kwargs['layers'] = [3, 4, 23, 3]
  kwargs['lconv'] = 3
  kwargs['block'] = Resnext
  return Resnet(**kwargs)

def resnext152(nc, **kwargs):
  kwargs['nc'] = nc
  kwargs['layers'] = [3, 8, 56, 3]
  kwargs['lconv'] = 3
  kwargs['block'] = ResneXt
  return Resnet(**kwargs)



"""
VGG Model
"""

import torch
import torch.nn as nn
from scratchai.nets.common import Flatten


__all__ = ['VGG', 'vgg11', 'vgg11_bn', 'vgg13', 'vgg13_bn', 'vgg16', 
           'vgg16_bn', 'vgg19', 'vgg19_bn']


def conv(ic:int, oc:int, k:int=3, s:int=1, p:int=1, norm:bool=True):
  layers = [nn.Conv2d(ic, oc, k, s, p, bias=not norm)]
  if norm: layers += [nn.BatchNorm2d(oc)]
  layers += [nn.ReLU(inplace=True)]
  return layers

def linear(inn:int, onn:int, drop:bool=True):
  layers = [nn.Linear(inn, onn)]
  if drop: layers += [nn.Dropout()]
  layers += [nn.ReLU(inplace=True)]
  return layers


# TODO Add the VGG A-LRN and VGG C models

class VGG(nn.Module):
  """
  Implementation of VGG model.
  Paper: https://arxiv.org/pdf/1409.1556.pdf

  Arguments
  ---------
  nc : int
       The number of classes
  lconf : list
          The configuration of the convs in the net, this is the conf
          that makes the implementation 11layers or 19layers or more as needed.
          Default: [1, 1, 2, 2, 2] - which is VGG11
  norm : bool
         If true, BatchNormalization is used after each layer.
         Defaults to True.
  """
  def __init__(self, nc=1000, lconf:list=[1, 1, 2, 2, 2], norm:bool=True):
    super().__init__()
    
    ic = 3; oc = 64
    layers = []
    for l in lconf:
      layers += conv(ic, oc, norm=norm)
      ic = oc
      for _ in range(l-1): layers += conv(ic, oc, norm=norm)
      oc *= 2 if oc*2 <= 512 else 1
      layers += [nn.MaxPool2d(2, 2)]
    
    layers += [nn.AdaptiveAvgPool2d((7, 7)), Flatten(), 
               *linear(512 * 7 * 7, 4096), *linear(4096, 4096), 
               nn.Linear(4096, nc)]

    self.net = nn.Sequential(*layers)

  def forward(self, x): return self.net(x)


### VGG A ###
def vgg11():
  return VGG(lconf=[1, 1, 2, 2, 2], norm=False)

def vgg11_bn():
  return VGG(lconf=[1, 1, 2, 2, 2])

### VGG B ###
def vgg13():
  return VGG(lconf=[2, 2, 2, 2, 2], norm=False)

def vgg13_bn():
  return VGG(lconf=[2, 2, 2, 2, 2])

### VGG D ###
def vgg16():
  return VGG(lconf=[2, 2, 2, 2, 2], norm=False)
 
def vgg16_bn():
  return VGG(lconf=[2, 2, 2, 2, 2])

### VGG E ###
def vgg19():
  return VGG(lconf=[2, 2, 4, 4, 4], norm=False)
  
def vgg19_bn():
  return VGG(lconf=[2, 2, 4, 4, 4])



import unittest
import torch
import torch.nn as nn
import PIL
import numpy as np
import scratchai

from PIL import Image
from scratchai import imgutils
from scratchai import *



#############################################
### Check functions for scratchai/imgutils.py
#############################################

class TestImgUtils(unittest.TestCase):
  
  url_1 = 'https://cdn.theatlantic.com/assets/media/img/mt/2015/10/' \
          '5594637726_3a0fdc9b7b_o/lead_720_405.jpg'

  def test_thresh_img(self):
    """
    This test makes sure that the thresh_img function 
    is working as expected.
    """
    noise = np.random.randint(0, 255, size=(32, 32, 3))
    thresh = [45, 45, 45]
    img = imgutils.thresh_img(noise, thresh, tcol=[255, 255, 255])

    self.assertEqual(type(img), np.ndarray, 'Return type not np.ndarray')
    self.assertEqual(img.shape, (32, 32, 3), 'The shape of output is' \
                          'not as expected')

    # TO make sure all pixels having value less than threshold is removed.
    # NOTE if tcol have values lower than thresh then values lower
    # than thresh will be present. That is why tcol = [255, 255, 255]
    # for the purposes of this test.
    self.assertFalse((img < thresh).any())

  def test_mask_reg(self):
    #TODO add tests for testing it further
    if not callable(getattr(imgutils, 'mask_reg', None)):
      raise NotImplementedError

  def test_mark_pnt_on_img(self):
    #TODO add tests for testing it further
    if not callable(getattr(imgutils, 'mark_pnt_on_img', None)):
      raise NotImplementedError
    
  def test_load_img(self):
    img = imgutils.load_img(TestImgUtils.url_1, np.ndarray)
    self.assertLessEqual(img.shape[-1], 3, '# of channels is not as expected')
    
  def test_t2i(self):
    noise = torch.randn(1, 3, 12, 12)
    image = imgutils.t2i(noise)
    self.assertEqual(type(image), PIL.Image.Image, 'Unexpected Type')
    image = np.array(image)
    self.assertEqual(list(image.shape), [12, 12, 3],'Unexpected shape')

  def test_imsave_imshow(self):
    #TODO add tests for testing it further
    utils.implemented(imgutils, 'imsave')
    utils.implemented(imgutils, 'imshow')

  def test_diff_imgs(self):
    t1 = torch.randn(3, 150, 150)
    t2 = torch.randn(3, 150, 150)
    self.assertTrue(imgutils.diff_imgs(t1, t1, normd=False).sum().item() == 0, 
               'nope!')
    self.assertTrue(imgutils.diff_imgs(t2, t2, normd=False).sum().item() == 0, 
               'nope!')
    self.assertFalse(imgutils.diff_imgs(t1, t2, normd=False).sum().item() == 0, 
               'nope!')
    
    # Explicit test
    t1 = torch.Tensor([[[1, 1, 1], [2, 2, 2], [3, 3, 3]]])
    t2 = torch.Tensor([[[2, 2, 2], [1, 1, 1], [4, 4, 4]]])
    t3 = torch.Tensor([[[-1, -1, -1], [1, 1, 1], [-1, -1, -1]]])
    t4 = torch.Tensor([[[1, 1, 1], [-1, -1, -1], [1, 1, 1]]])

    # With norm false
    dimg1 = imgutils.diff_imgs(t1, t2, normd=False)
    dimg2 = imgutils.diff_imgs(t2, t1, normd=False)
    self.assertTrue((dimg1 - t3.squeeze()).sum().item() == 0., 'nope!')
    self.assertTrue((dimg2 - t4.squeeze()).sum().item() == 0., 'nope!')

  def test_mean(self):
    for _ in range(torch.randint(1, 10, ())):
      t = torch.randint(0, 255, size=(3, 28, 28))
      tm = t.sum() / t.numel()
      m = imgutils.mean(t)
      self.assertEqual(tm, m, 'mean not correct!')
    
  def test_std(self):
    for _ in range(torch.randint(1, 10, ())):
      t = torch.randn(3, 28, 28)
      std = imgutils.std(t)
      self.assertEqual(round(std - torch.std(t).item(), 2), 0.0, 'nope!')


#############################################
### Check the functions in scratchai/utils.py
#############################################

class TestUtils(unittest.TestCase):
  
  def test_load_from_pth(self):
    #TODO add tests for testing it further
    if not callable(getattr(scratchai.utils, 'load_from_pth', None)):
      raise NotImplementedError

  def test_name_from_object(self):
    obj = nets.Lenet()
    name = utils.name_from_object(obj)
    self.assertTrue(name == 'lenet', 'doesn\t look good')
  
  def test_freeze(self):
    net = nets.resnet18()
    for p in net.parameters():
      self.assertTrue(p.requires_grad, 'initialization already frozen!')
    utils.freeze(net)
    for p in net.parameters():
      self.assertFalse(p.requires_grad, 'not working!')

  def test_topk(self):
    name = 'Acc'; cnt = 50
    for _ in range(torch.randint(1, 10, ())):
      topk = tuple(torch.randint(2, 10, size=(5,)).numpy())
      obj = utils.Topk(name, topk)
      topk = sorted(tuple(set(topk)))
      val = np.zeros(len(topk))

      for i in range(1, torch.randint(1, 10, size=())):
        curr_val = np.random.randint(0, 50, size=(len(topk),)).astype('float')
        obj.update(curr_val/cnt, cnt)
        val += curr_val
        k = np.random.choice(topk)
        self.assertEqual(round(obj.avgmtrs[name+str(k)].avg - 
                         val[topk.index(k)]/(cnt*i), 2), 0.0, 'nope!')
        
      self.assertEqual(len(obj.avgmtrs), len(topk), 'not working!')
      self.assertEqual(obj.ks, len(topk), 'not working!')
      # If len(topk) < 2: then a 0-d tensor will be passed which will throw exp
      if len(topk) > 2: self.assertRaises(AssertionError, 
                        lambda: obj.update(val[:len(topk)-1], cnt))
      
    topk = tuple(torch.randint(1, 10, size=(5,)).numpy())
    obj = utils.Topk(name, topk)
    self.assertEqual(name, obj.name, 'dude!')
    self.assertRaises(AssertionError, lambda: utils.Topk('A', (1, 3, 0)))

  def test_avgmeter(self):
    name = 'name'; fmt = '.:2f'
    mtr = utils.AvgMeter(name, fmt)
    self.assertEqual(mtr.name, name, 'initialization went wrong!')
    self.assertEqual(mtr.fmt, fmt, 'initialization went wrong!')
    self.assertEqual(mtr.val + mtr.sum + mtr.cnt + mtr.avg, 0, 'result bad!')
    
    val = 0.; cnt = 0.
    for _ in range(torch.randint(1, 10, ())):
      curr_val = torch.randint(1, 10, ()).float()
      curr_cnt = torch.randint(1, 10, ()).float()
      mtr(curr_val/curr_cnt, curr_cnt)
      val += curr_val; cnt += curr_cnt
      self.assertEqual(mtr.avg, (val / cnt), 'result bad!')

    self.assertEqual(str(mtr), '{} - {}'.format(name, val/cnt))

  def test_setatrib(self):
    # TODO Needs more tests
    net = nets.alexnet()
    val = nn.Identity()
    t = net.net[0]; utils.setatrib(net, 'net[0]', val)
    self.assertEqual(net.net[0], val, 'not working!')
    self.assertFalse(t == val, 'not working!')
  
  def test_count_params(self):
    net = nn.Conv2d(3, 4, 5, 2, 1)
    num = utils.count_params(net)
    self.assertEqual(num, (4*5*5*3) + 4, 'nope!')

  def test_sgdivisor(self):
    ulim = 100
    for n in np.random.randint(1, ulim, size=(5,)):
      s, g = utils.sgdivisor(int(n))
      self.assertTrue(n % s == 0, 'Nope!')
      self.assertTrue(n % g == 0, 'Nope!')
      # Explicit checks
      for ii in range(s-1, 1, -1):
        if ii == n: continue
        self.assertFalse(n % ii == 0, 'nope!')
      for ii in range(g+1, ulim):
        if ii == n: continue
        self.assertFalse(n % ii == 0, 'nope!')

  def test_gpfactor(self):
    pfs = [2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,
           79,83,89,97]
    # DO NOT CHANGE ulim AS pfs is a list of prime numbers b/w # 0 and 100.
    ulim = 100
    for n in np.random.randint(1, ulim, size=(5,)):
      gp = utils.gpfactor(int(n))
      self.assertTrue(n % gp == 0, 'Nope!')
      # Explicit checks
      for pf in pfs:
        if n % pf == 0: self.assertFalse(pf > gp , 'nope!')
    self.assertRaises(AssertionError, lambda: utils.gpfactor(0))

#############################################
### Check the functions in scratchai/attacks/utils.py
#############################################

class TestAtkUtils(unittest.TestCase):
  
  def test_optimize_linear(self):
    #TODO add tests for testing it further
    if not callable(getattr(scratchai.attacks.utils, 'optimize_linear', None)):
      raise NotImplementedError

  def test_clip_eta(self):
    #TODO add tests for testing it further
    if not callable(getattr(scratchai.attacks.utils, 'clip_eta', None)):
      raise NotImplementedError



import torch
import torch.nn as nn

def bnconv(in_channels:int, out_channels:int, kernel_size, stride,
             padding:int, bias:bool=False, eps:float=1e-5, 
             momentum:float=0.1, conv_first=True, act:str='relu', 
             nslope:float=0.2, norm:str='batch', inplace:bool=True):
    
    '''
    This is the BNConv module that helps in defining the 
    Conv->Norm->RelU
    RelU->Norm->Conv
    
    Arguments:
    - in_channels = # of input channels
    - out_channels = # of output channels
    - kernel_size = the kernel_size
    - stride - the stride
    - padding - the padding
    - bias = whether to add bias to Convolutional Layer
    - eps = the epsilon value
    - momentum = the momentum value
    - conv_first = If True: conv->norm->act
                   If False: act->norm->conv
                   Default: True
    - nslope = the slope if leaky relu is used.
    - norm = the norm type
    - inplace = The value for the inplace argument inactivation
                Default: True
    '''

    normc = out_channels if conv_first else in_channels
    
    if act == 'relu':
        act = nn.ReLU(inplace=inplace)
    elif act == 'leakyrelu':
        act = nn.LeakyReLU(negative_slope=nslope, inplace=inplace)
    elif act == False:
        act = None
    else:
        raise ('Activation value not understood')

    if norm == 'batch':
        norm = nn.BatchNorm2d(normc, eps=eps, momentum=momentum)
    elif norm == 'instance':
        norm = nn.InstanceNorm2d(normc, eps=eps, momentum=momentum)
    else:
        raise ('Norm value not understood')
    
    layers = [nn.Conv2d(in_channels=in_channels,
                        out_channels=out_channels,
                        kernel_size=kernel_size,
                        padding=padding,
                        stride=stride,
                        bias=bias),
                
              norm,
              act]
    
    if act is None:
        layers = layers[:2]

    net = nn.Sequential(*layers) if conv_first else nn.Sequential(*layers[::-1])

    return net



from scratchai import learners
from scratchai import DataLoader
from scratchai import utils
from scratchai import nets
from scratchai import init
from scratchai import attacks
from scratchai import pretrained
from scratchai import imgutils
from scratchai import one_call
from scratchai._config import *

import os

# TODO Confirm if this is the most efficient way to do so.
if not os.path.exists(home):
  os.makedirs(home)



"""
The Projected Gradient Descent attack.
"""

import numpy as np
import torch
import torch.nn as nn
from scratchai.attacks.attacks import fgm
from scratchai.attacks.utils import clip_eta


__all__ = ['pgd', 'PGD']


def pgd(x:torch.Tensor, net:nn.Module, nb_iter:int=10, eps:float=0.3, 
    eps_iter:float=0.05, rand_minmax:float=0.3, clip_min=None, clip_max=None, 
    y=None, ordr=np.inf, rand_init=None, targeted=False) -> torch.Tensor:

  """
  This class implements either the Basic Iterative Method
  (Kuarkin et al. 2016) when rand_init is set to 0. or the
  Madry et al. (2017) method when rand_minmax is larger than 0.
  Paper link (Kuarkin et al. 2016): https://arxiv.org/pdf/1607.02533.pdf
  Paper link (Madry et al. 2017): https://arxiv.org/pdf/1706.06083.pdf
 
  # TODO FIX the below arguments style
  Arguments
  ---------
  model: Model
  dtype: dtype of the data
  default_rand_init: whether to use random initialization by default
  kwargs: passed through to super constructor

  Returns
  -------
  adv_x : torch.Tensor
      The adversarial Example.
  """
  # TODO Check params
  # If a data range was specified, check that the input was in that range
  if clip_min is not None:
    asserts.append(x.any() >= clip_min)

  if clip_max is not None:
    asserts.append(x.any() <= clip_max)

  # Initialize loop variables
  if rand_init:
    eta = torch.FloatTensor(*x.shape).uniform_(-minmax, minmax)
  else:
    eta = torch.zeros_like(x)

  # Clip eta
  eta = clip_eta(eta, ordr, eps)
  adv_x = x + eta
  
  if clip_min is not None or clip_max is not None:
    adv_x = torch.clamp(adv_x, clip_max, clip_min)
  
  if y is None:
    # Use ground truth labels to avoid label leaking
    if len(x.shape) == 3: x.unsqueeze_(0)
    _, y = torch.max(net(x), dim=1)
  else:
    targeted = True
 
  if ord == 1:
    raise NotImplementedError("It's not clear that FGM is a good inner loop"
                 " step for PGD when ord=1, because ord=1 FGM "
                 " changes only one pixel at a time. We need "
                 " to rigoursly test a strong ord=1 PGD "
                 " before enabling this feature.")
  i = 0
  while i < nb_iter:
    """
    Do a projected gradient step.
    """
    adv_x = fgm(adv_x, net, eps=eps_iter, ordr=ordr, clip_min=clip_min,
                clip_max=clip_max, y=y, targeted=targeted)

    # Clipping perturbation eta to ord norm ball
    eta = adv_x - x
    eta = clip_eta(eta, ordr, eps)
    adv_x = x + eta

    # Redo the clipping.
    # FGM alread already did it, but subtracting and re-adding eta can add some
    # small numerical error
    if clip_min is not None or clip_max is not None:
      adv_x = torch.clamp(adv_x, clip_min, clip_max)
    i += 1

  # Asserts run only on CPU
  # When multi-GPU eval code tries to force all PGD ops onto GPU, this
  # can cause an error.
  # The 1e-6 is needed to compensate for numerical error.
  # Without the 1e-6 this fails when e.g. eps=.2 clip_max=.5 clip_min=.7
  if ordr == np.inf and clip_min is not None:
    assert (eps <= (1e6 + clip_max - clip_min)) 
  
  return adv_x


##########################################################################
###### Class to initialize the attack for use with torchvision.transforms

class PGD():
  def __init__(self, net, **kwargs):
    self.net = net
    self.kwargs = kwargs
  def __call__(self, x):
    return pgd(x, self.net, **self.kwargs)



import scratchai
import torch
import torch.nn as nn
import unittest
import requests
import zipfile
import io
import numpy as np
from torchvision import models, transforms
from PIL import Image
from inspect import isfunction
import matplotlib.pyplot as plt

from scratchai import *

NOISE = 'noise'
SEMANTIC = 'semantic'
FGM = 'fgm'
PGD = 'pgd'
DEEPFOOL = 'deepfool'

class TestAttacks(unittest.TestCase):
  
  # TODO Shorten the url
  url = 'https://www.publicdomainpictures.net/pictures/210000/nahled/tiger-in-the-water-14812069667ks.jpg'
  trf = imgutils.get_trf('rz256_cc224_tt_normimgnet')
  
  url_dset = 'https://www.dropbox.com/s/6bg8ntqcs4r98i9/testdataset.zip?dl=1'

  def test_noise_atk(self):
    """
    Tests to check that the Noise Attack works
    """
    with open('/tmp/test.png', 'wb') as f:
      f.write(requests.get(TestAttacks.url).content)
    img = Image.open('/tmp/test.png')

    #all_models = ['alexnet', 'resnet18', 'resnet34', 'resnet50', 'resnet101', 'resnet152']
    all_models = ['alexnet']
    for model in all_models:
      print ('[INFO] Testing Noise attack on {}'.format(model))
      net = getattr(nets, model)(pretrained=True)
      # TODO No need to call Noise again and again in each iteration
      self.check_atk(net, img, attacks.noise, t=NOISE)
      self.check_atk(net, img, attacks.Noise(), t=NOISE)
      print ('[INFO] Attack worked successfully!')
      del net

  def test_semantic(self):
    """
    tests to ensure semantic attack works!
    """
    with open('/tmp/test.png', 'wb') as f:
      f.write(requests.get(TestAttacks.url).content)
    img = Image.open('/tmp/test.png')

    # Maybe an option to perform the rigourous testing, if needed.
    #all_models = ['alexnet', 'resnet18', 'resnet34', 'resnet50', 'resnet101', 'resnet152']
    all_models = ['alexnet']
    for model in all_models:
      print ('[INFO] Testing Semantic attack on {}'.format(model))
      net = getattr(nets, model)(pretrained=True)
      # TODO No need to call Noise again and again in each iteration
      self.check_atk(net, img, attacks.semantic, t=SEMANTIC)
      self.check_atk(net, img, attacks.Semantic(), t=SEMANTIC)
      print ('[INFO] Attack worked successfully!')
      del net
  
  def test_pgd(self):
    """
    tests to ensure that fast_gradient_method attack works
    """
    with open('/tmp/test.png', 'wb') as f:
      f.write(requests.get(TestAttacks.url).content)
    img = Image.open('/tmp/test.png')

    # TODO replace this with a scratchai model
    # Running inference on just alexnet as it takes too long, otherwise
    # TODO Update other tests to just infer with one model for quick testing and 
    # Maybe an option to perform the rigourous testing, if needed.
    #all_models = ['alexnet', 'resnet18', 'resnet34', 'resnet50', 'resnet101', 'resnet152']
    all_models = ['alexnet']
    for model in all_models:
      to_pred = int(torch.randint(1000, ()))
      print ('[INFO] Testing PGD attack on {}'.format(model))
      net = getattr(models, model)(pretrained=True)
      self.check_atk(net, img, attacks.pgd, t=PGD, y=torch.tensor([to_pred]))
      self.check_atk(net, img, attacks.PGD(net,  y=torch.tensor([to_pred])),
                      t=PGD)
      print ('[INFO] Attack worked successfully!')
      del net

  def test_fgm(self):
    """
    tests to ensure that fast_gradient_method attack works
    """
    with open('/tmp/test.png', 'wb') as f:
      f.write(requests.get(TestAttacks.url).content)
    img = Image.open('/tmp/test.png')

    # TODO replace this with a scratchai model
    # TODO write tests for targeted attack with fgm
    # TODO Update other tests to just infer with one model for quick testing and 
    # Maybe an option to perform the rigourous testing, if needed.
    #all_models = ['alexnet', 'resnet18', 'resnet34', 'resnet50', 'resnet101', 'resnet152']
    all_models = ['alexnet']
    for model in all_models:
      print ('[INFO] Testing FGM attack on {}'.format(model))
      net = getattr(models, model)(pretrained=True)
      self.check_atk(net, img, scratchai.attacks.fgm, t=FGM)
      self.check_atk(net, img, scratchai.attacks.FGM(net), t=FGM)
      print ('[INFO] Attack worked successfully!')
      del net
  
  def test_deepfool(self):
    """
    tests to ensure that DeepFool attack works.
    """
    with open('/tmp/test.png', 'wb') as f:
      f.write(requests.get(TestAttacks.url).content)
    img = Image.open('/tmp/test.png')

    all_models = ['alexnet']
    for model in all_models:
      print ('[INFO] Testing DeepFool attack on {}'.format(model))
      net = getattr(nets, model)(pretrained=True)
      self.check_atk(net, img, attacks.deepfool, t=DEEPFOOL)
      self.check_atk(net, img, attacks.DeepFool(net), t=DEEPFOOL)
      print ('[INFO] Attack worked successfully!')
      del net
    
  def test_benchmark(self):
    net = nets.resnet18()
    if not os.path.exists('/tmp/imgnet/'):
      r = requests.get(TestAttacks.url_dset)
      z = zipfile.ZipFile(io.BytesIO(r.content))
      z.extractall('/tmp/')

    #atks = [attacks.Noise, attacks.Semantic, attacks.FGM, attacks.PGD, 
            #attacks.DeepFool]
    atks = [attacks.Noise]
    for atk in atks: attacks.benchmark_atk(atk, net, root='/tmp/imgnet/')

  def scale(self, img):
    return img * (255. / img.max())

  def check_atk(self, net, img, atk, t, y=None):
    """
    Arguments
    ---------
    init : bool
           If set to True, the funtion assumes that the attack passed is 
           initialized from the attack class i.e.
           atk = attacks.PGD(net)
           else if False it assumes
           atk = attacks.noise
           Defaults to False.
    """
    init = not isfunction(atk)

    # Get true pred
    net.eval()
    utils.freeze(net)
    true_pred = int(torch.argmax(net(TestAttacks.trf(img).unsqueeze(0)), dim=1))
    
    # Adversarial Example
    if t == NOISE:
      adv_x = atk(torch.from_numpy(np.array(img)).float()).transpose(2, 1).transpose(1, 0)
      adv_pred = int(torch.argmax(net(adv_x.unsqueeze(0)), dim=1))
    elif t == SEMANTIC:
      img = TestAttacks.trf(img)
      #img = self.scale(img) # It works w/ as well as w/o scaling.
      adv_x = atk(img)
      #plt.imshow(adv_x.transpose(0, 1).transpose(1, 2)); plt.show()
      adv_x = TestAttacks.trf(transforms.ToPILImage()(adv_x))
      adv_pred = int(torch.argmax(net(adv_x.unsqueeze(0)), dim=1))
    elif t == FGM or t == PGD:
      img = TestAttacks.trf(img)
      if init: adv_x = atk(img.unsqueeze(0))
      else: adv_x = atk(img.unsqueeze(0), net, y=y)
      adv_pred = int(torch.argmax(net(adv_x), dim=1))
    elif t == DEEPFOOL:
      img = TestAttacks.trf(img)
      if init: adv_x = atk(img.unsqueeze(0))
      else: adv_x = atk(img.unsqueeze(0), net)
      adv_pred = int(torch.argmax(net(adv_x), dim=1))
      
    print (true_pred, adv_pred)
    self.assertFalse(true_pred == adv_pred, 'The attack doesn\'t work!')
    if y is not None:
      self.assertTrue(adv_pred == int(y), 'The attack doesn\'t work!')
      print ('TARGETED ATTACK WORKS!')



"""
Wrappers to quickly train on common datasets.
"""

import numpy as np
import matplotlib.pyplot as plt

from scratchai.imgutils import get_trf
from scratchai.learners.clflearner import *
from scratchai._config import home
from scratchai import utils
from scratchai._config import CIFAR10, MNIST


def mnist(net, **kwargs):
  """
  Train on MNIST with net.

  Arguments
  ---------
  net : nn.Module
        The net which to train.
  optim : nn.optim
          The optimizer to use.
  crit : nn.Module
         The criterion to use.
  lr : float
       The learning rate.
  wd : float
       The weight decay.
  bs : int
       The batch size.
  seed : int
         The particular seed to use.
  epochs : int
           The epcochs to train for.
  ckpt : str
         Path to the ckpt file.
  resume : bool
           If true, resume training from ckpt.
           else not.
  root : str
         The root where the datasets is or
         needs to be downloaded.

  Returns
  -------
  tlist : list
          Contains list of n 2-tuples. where n == epochs
          and a tuple (a, b) where,
          a -> is the acc for the corresponding index
          b -> is the loss for the corresponding index
          for training
  vlist : list
          Contains list of n 2-tuples. where n == epochs
          and a tuple (a, b) where,
          a -> is the acc for the corresponding index
          b -> is the loss for the corresponding index
          for validation
  """
  
  opti, crit, kwargs = preprocess_opts(net, dset=MNIST, **kwargs)

  trf = get_trf('rr20_tt_normmnist')

  t = datasets.MNIST(kwargs['root'], train=True, download=True, transform=trf)
  v = datasets.MNIST(kwargs['root'], train=False, download=True, transform=trf)
  tloader = DataLoader(t, shuffle=True, batch_size=kwargs['bs'])
  vloader = DataLoader(v, shuffle=True, batch_size=kwargs['bs'])
  
  tlist, vlist = clf_fit(net, crit, opti, tloader, vloader, **kwargs)
  plt_tr_vs_tt(tlist, vlist)


def cifar10(net, **kwargs):
  """
  Train on CIFAR10 with net.

  Arguments
  ---------
  net : nn.Module
        The net which to train.
  optim : nn.optim
          The optimizer to use.
  crit : nn.Module
         The criterion to use.
  lr : float
       The learning rate.
  wd : float
       The weight decay.
  bs : int
       The batch size.
  seed : int
         The particular seed to use.
  epochs : int
           The epcochs to train for.
  ckpt : str, None
         Path to the ckpt file. If not None, training is started
         using this ckpt file. Defaults to None
  root : str
         The root where the datasets is or
         needs to be downloaded.
  
  Returns
  -------
  tlist : list
          Contains list of n 2-tuples. where n == epochs
          and a tuple (a, b) where,
          a -> is the acc for the corresponding index
          b -> is the loss for the corresponding index
          for training
  vlist : list
          Contains list of n 2-tuples. where n == epochs
          and a tuple (a, b) where,
          a -> is the acc for the corresponding index
          b -> is the loss for the corresponding index
          for validation
  """
  opti, crit, kwargs = preprocess_opts(net, dset=CIFAR10, **kwargs)

  trf = get_trf('pad4_rc32_tt_normimgnet')

  t = datasets.CIFAR10(kwargs['root'], train=True, download=True, transform=trf)
  v = datasets.CIFAR10(kwargs['root'], train=False, download=True, transform=trf)
  tloader = DataLoader(t, shuffle=True, batch_size=kwargs['bs'])
  vloader = DataLoader(v, shuffle=True, batch_size=kwargs['bs'])
  
  tlist, vlist = clf_fit(net, crit, opti, tloader, vloader, **kwargs)
  plt_tr_vs_tt(tlist, vlist)


def custom(net, tloader, vloader, **kwargs):
  """
  Train on a custom dataset with net.

  Arguments
  ---------
  net : nn.Module
        The net which to train.
  optim : nn.optim
          The optimizer to use.
  crit : nn.Module
         The criterion to use.
  lr : float
       The learning rate.
  wd : float
       The weight decay.
  bs : int
       The batch size.
  seed : int
         The particular seed to use.
  epochs : int
           The epcochs to train for.
  ckpt : str, None
         Path to the ckpt file. If not None, training is started
         using this ckpt file. Defaults to None
  root : str
         The root where the datasets is or
         needs to be downloaded.
  
  Returns
  -------
  tlist : list
          Contains list of n 2-tuples. where n == epochs
          and a tuple (a, b) where,
          a -> is the acc for the corresponding index
          b -> is the loss for the corresponding index
          for training
  vlist : list
          Contains list of n 2-tuples. where n == epochs
          and a tuple (a, b) where,
          a -> is the acc for the corresponding index
          b -> is the loss for the corresponding index
          for validation
  """
  opti, crit, kwargs = preprocess_opts(net, **kwargs)

  #trf = get_trf('rz256_cc224_tt_normimgnet')
  trf = get_trf('rz32_tt_normimgnet')

  tlist, vlist = clf_fit(net, crit, opti, tloader, vloader, **kwargs)
  plt_tr_vs_tt(tlist, vlist)


def preprocess_opts(net, dset:str=None, **kwargs):
  """
  Helper function that abstracts away the preprocessing of the
  default kwargs as required for each dataset.

  Arguments
  ---------
  net : nn.Module
        The net to train.
  dset : str, None
         Name of the dataset.
  kwargs : dict
           The dict with all the keys and values.
  Returns
  -------
  dict : kwargs
         The passed dict with all the values.
  """

  if 'optim' not in kwargs: kwargs['optim'] = optim.SGD
  if 'crit' not in kwargs: kwargs['crit'] = nn.CrossEntropyLoss
  if 'lr' not in kwargs: kwargs['lr'] = 3e-4
  if 'wd' not in kwargs: kwargs['wd'] = 0
  if 'mom' not in kwargs: kwargs['mom'] = 0.9
  if 'nestv' not in kwargs: kwargs['nestv'] = False
  if 'bs' not in kwargs: kwargs['bs'] = 16
  if 'seed' not in kwargs: kwargs['seed'] = 123
  if 'epochs' not in kwargs: kwargs['epochs'] = 5
  if 'topk' not in kwargs: kwargs['topk'] = (1, 5)
  if 'lr_step' not in kwargs: kwargs['lr_step'] = None
  if 'lr_decay' not in kwargs: kwargs['lr_decay'] = 0.2
  if 'ckpt' not in kwargs: kwargs['ckpt'] = None
  if 'root' not in kwargs: kwargs['root'] = home
  
  # Set Dataset specific values if dset is not None here

  for key, val in kwargs.items():
    print ('[INFO] Setting {} to {}.'.format(key, val))
  
  lr = kwargs['lr']
  wd = kwargs['wd']
  mom = kwargs['mom']

  crit = kwargs['crit']()
  opti_name = utils.name_from_object(kwargs['optim'])
  train_params = [p for p in net.parameters() if p.requires_grad]
  if  opti_name == 'adam':
    opti = kwargs['optim'](train_params, lr=lr, weight_decay=wd)
  elif opti_name == 'sgd':
    opti = kwargs['optim'](train_params, lr=lr, nesterov=kwargs['nestv'],
                           weight_decay=wd, momentum=mom)
  else:
    raise NotImplementedError

  # Resume from ckpt (if ckpt is not None)
  if kwargs['ckpt'] is not None:
    ckpt = torch.load(kwargs['ckpt'])
    print ('[INFO] Looking for key "opti" in ckpt file...')
    opti.load_state_dict(ckpt['opti'])
    # If optimizer is SGD, then momentum buffers needs to be moved to device
    # See: https://discuss.pytorch.org/t/runtimeerror-expected-type-torch-
    # floattensor-but-got-torch-cuda-floattensor-while-resuming-training/37936
    if opti_name == 'sgd':
      for p in opti.state.keys():
        buf = opti.state[p]['momentum_buffer']
        opti.state[p]['momentum_buffer'] = buf.cuda()
    print ('[INFO] Found and loaded the optimizer state_dict')
    print ('[INFO] Looking for key "net" in ckpt file...')
    net.load_state_dict(ckpt['net'])
    print ('[INFO] Found and loaded the model state_dict')

  # Pop keys from kwargs to avoid 
  # TypeError: got multiple values for 1 argument
  kwargs.pop('crit', None); kwargs.pop('opti', None); 

  return opti, crit, kwargs


def plt_tr_vs_tt(tlist, vlist):
  tacc = list(map(lambda x : x[0][0], tlist))
  tloss = list(map(lambda x : x[1], tlist))
  vacc = list(map(lambda x : x[0][0], vlist))
  vloss = list(map(lambda x : x[1], vlist))
  epochs = np.arange(1, len(tlist)+1)
  plt.plot(epochs, tacc, 'b--', label='Train Accuracy')
  plt.plot(epochs, vacc, 'b-', label='Val Accuracy')
  plt.plot(epochs, tloss, 'o--', label='Train Loss')
  plt.plot(epochs, vloss, 'o-', label='Val Loss')
  plt.xlabel('Epochs')
  plt.legend()
  plt.show()



from scratchai.nets.seg.enet import *
from scratchai.nets.seg.unet import *

from scratchai.nets.clf.resnet import *
from scratchai.nets.clf.vgg import *
from scratchai.nets.clf.lenet import *
from scratchai.nets.clf.alexnet import *

from scratchai.nets.style_transfer import *

from scratchai.nets.common import *



"""
Contains some commonly used functions b/w all modules.
"""

import torch
import torch.nn as nn

class Flatten(nn.Module):
  def forward(self, x):
    return x.view(x.size(0), -1)


class Debug(nn.Module):
  def __init__(self, debug_value=None):
    super().__init__()
    self.debug_value = debug_value
  def forward(self, x):
    if self.debug_value: print (self.debug_value, end='')
    print (x.shape)
    return x



resnet18_url = 'https://www.dropbox.com/s/ln4i4547jvn5wr8/resnet18.pth?dl=1'
resnet34_url = 'https://www.dropbox.com/s/2mxj66dqtfqss2c/resnet34.pth?dl=1'
resnet50_url = 'https://www.dropbox.com/s/hinaadj4kj1rsl0/resnet50.pth?dl=1'
resnet101_url = 'https://www.dropbox.com/s/11u40th2ygsz9u0/resnet101.pth?dl=1'
resnet152_url = 'https://www.dropbox.com/s/9babg74hcgx1w9l/resnet152.pth?dl=1'
elephant_skin_url = 'https://www.dropbox.com/s/xak3qtgy426nwig/elephant_skin%281%29.pth?dl=1'
snake_skin_url = 'https://www.dropbox.com/s/98nmx8gcsce49zr/snake_skin.pth?dl=1'
lenet_mnist_url = 'https://www.dropbox.com/s/0ylrj6a8gx1iiuw/lenet_mnist.pth?dl=1'
alexnet_url = 'https://www.dropbox.com/s/hogfy6yap8an5x9/alexnet.pth?dl=1'
alexnet_mnist_url = 'https://www.dropbox.com/s/u123zs9t6rcco18/alexnet_mnist.pth?dl=1'

# This is a test dataset which is very small in size and 
# helps ensures things are working as they are supposed to
tdset = 'https://www.dropbox.com/s/79cnd89nxjm60u6/testdataset.zip?dl=1'



"""
The Saliency Map Method Attack
"""

import numpy as np
import torch
import torch.nn as nn
from scratchai.attacks.attacks.attack import Attack

class SaliencyMapMethod(Attack):
    """
    The Jacobian-based Saliency Map Method (Papernot et al. 2016)
    Paper Link: https://arxiv.org/pdf/1511.07528.pdf

    Arguments
    ---------
    model : nn.Module
            The model on which the attack needs to be performed.
    dtype : str
            The data type of the model.

    Returns
    -------
    adv : torch.tensor
          The adversarial Example of the input.
    """

    def __init__(self, model, dtype='float32', **kwargs):
        super().__init__(model, dtype, **kwargs)

    def generate(self, x, **kwargs):
        """
        Generate symbolic graph for adversarial examples and return.

        Arguments
        ---------
        x : torch.tensor
            The input to the model
        kwargs : dict
                 Additonal arguments
        """

        # Parse and save attack specific parameters
        assert self.parse_params(**kwargs)

        if self.y_target is None:
            def random_targets(gt):
                result = gt.copy()
                nb_s = gt.shape[0]
                nb_classes = gt.shape[1]

                for i in range(nb_s):
                    result[i, :] = np.roll(result[i, :],
                                   np.random.randint(1, nb_classes-1)
                return result
            
            labels, nb_classes = self.get_or_guess_labels(x, kwargs)
            self.y_target = random_targets(labels)
            self.y_target = self.y_target.view([None, nb_classes])
        
        x_adv = x
        # TODO Uncomment and check
        #x_adv = jsma_symbolic(x, y_target, model, theta, gamma, clip_min,
                              #clip_max)
        return x_adv

    def parse_params(self, theta=1., gamma=1., clip_min=0., clip_max=1.,
                     y_target=None):
        """
        Takes in a dictionary of parameters and applies attack-specific checks
        before saving them as attributes.

        Attack-specific parameters:
        ---------------------------
        theta : float, optional
                Perturbation introduced to modified components
                (can be positive or negative). Defaults to 1.
        gamma : float, optional
                Maximum percentage of perturbed features. Defaults to 1.
        clip_min : float, optional
                   Minimum component value for clipping
        clip_max : float, optional
                   Maximum component value for clipping
        y_target : torch.tensor, optional
                   Target tensor if the attack is targetted
        """

        self.theta = theta
        self.gamma = gamma
        self.clip_min = clip_min
        self.clip_max = clip_max
        self.y_target = y_target
        
        return True


def jsma_symbolic(x, y_target, model, theta, gamma, clip_min, clip_max):
    """
    PyTorch Implementation of the JSMA (see https://arxiv.org/abs/1511.07520
    for the details about the algorithm design choices).

    Arguments
    ---------
    x : torch.tensor
        The input to the model
    y_target : torch.tensor
               The target tensor
    model : nn.Module
            The pytorch model
    theta : float
            delta for each feature adjustment.
    gamma : float
            a float between 0 and 1 indicating the maximum distortion
            percentage.
    clip_min : float
               minimum value for components of the example returned
    clip_max : float
               maximum value for components of the example returned.
    
    Returns
    -------
    x_adv : torch.tensor
            The adversarial example.
    """

    classes = int(y_target.shape[1])
    features = int(np.product(x.shape[1:]))

    max_iters = np.floor(features * gamma / 2)
    increase = bool(theta > 0)

    tmp = np.ones((features, features), int)
    np.fill_diagonal(tmp, 0)

    # Compute the initial search domain. We optimize the initial search domain
    # by removing all features that are already at their maximum values
    # (if increasing input features -- otherwise, at their minimum value).
    if increase:
        search_domain = (x < clip_max).view(-1, features)
    else:
        search_domain = (x > clip_min).view(-1, features)



import numpy as np
from PIL import Image
from scratchai.DataLoader.DatasetLoader import DatasetLoader
import matplotlib.pyplot as plt
import torch
import torchvision.transforms as trfs

class ImageLoader(DatasetLoader):
    '''
    This is the class that helps you load inputs and labels for semantic
    segmentation datasets.

    '''

    def __init__(self, input_path, label_path):
        '''
        The constructor to the SegmentationDatasetLoader class

        Arguments:
        - input_path : The path to the directory for input images
        - label_path : The path to the directory for label images
        '''

        super().__init__(input_path, label_path)


    def get_batch(self, shuffle=True):
        '''
        Return a batch of the images from both inputs and labels

        Arguments:
        - sample_size = The number of images to return

        returns;
        - 2 - np.ndarray : (N x C x H x W)
                       where,
                       N - is the sample_size
                       C - the number of channels
                       H - the height of the image
                       W - the width of the image

        '''
        
        if shuffle:
            sample_idx = np.random.randint(low=0, 
                                   high=self.tinp, 
                                   size=self.bs)
        else:
            sample_idx = self.glowhigh(batch_size=self.bs,
                                  return_range=True)
            
        inputs = self.inpn[sample_idx]
        labels = self.labn[sample_idx]
        
        all_inps = []
        all_labs = []
        for inp, label in zip(inputs, labels):
            si = np.array(Image.open(self.input_path + inp))
            si = torch.tensor(si).transpose(2, 1).transpose(1, 0)

            li = np.array(Image.open(self.label_path + label))
            li = torch.tensor(li)

            all_inps.append(si)
            all_labs.append(li)

        self.x = torch.stack(all_inps, dim=0)
        self.y = torch.stack(all_labs, dim=0).squeeze()
        
        yield self.x, self.y
            
    def create_loader(self, batch_size=1, shuffle=True):
        while(1):
            yield self.get_batch(batch_size=batch_size,
                                 shuffle=True)

    def show_few(self, size=1, in_row=False, figsize=(20, 10), show_axis=False):

        # Check argument data types
        if not isinstance(show_axis, bool):
            raise RuntimeError('Unexpected data type passed for "show_axis" argument!')
        if not isinstance(in_row, bool):
            raise RuntimeError('Unexpected data type passed for "in_row" argument!')

        inputs, labels = self.get_batch(batch_size=size)
        
        rc_tuple = (size, 1)
        if in_row:
            rc_tuple = (1, size)

        if not show_axis:
            show_axis = 'off'
        else:
            show_axis = 'on'

        figure = plt.subplots(figsize=figsize)
        for ii in range(1, size+1):
            plt.subplot(*rc_tuple, ii)
            plt.imshow(inputs[ii-1])

        plt.show()



import numpy as np
import os
from scratchai.DataLoader.ImageLoader import ImageLoader
from . import color_code as colors
from PIL import Image
import glob
import torchvision
from torch.utils.data import DataLoader
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import torchvision.transforms as trf
import cv2
import os
import subprocess

__all__ = ['SegLoader', 'camvid']

class SegLoader(ImageLoader):

  def __init__(self, ip:str, lp:str, nc:int, bs:int, trfs=None, imdf=True, d=None, cmap=None):
    """
    Constructor for the Segmentation Dataset Loader.

    Arguemnts
    ---------
    ip : str
         The input path
    lp : str
         The label path
    bs : int
         The batch size
    trfs : torchvision.transforms
           The transforms that needs to be performed on the Images
    imdf : bool
           If the image files are immediately in the paths mentioned
    dataset_is : str
                name of the dataset, if it is known then the color map will
                be loaded by default, without the need of passing
                Supported Datasets:
                  - CamVid : pass 'camvid' as value to this argument

    color_map : dict 
                a dictionary where the key is the class name
                and the value is a tuple or list with 3 elements
                one for each channel. So each key is a RGB value.
    """

    super().__init__(ip, lp)
    
    self.d = d
    if str(self.d) == colors.CAMVID:
      self.cmap = colors.camvid_color_map

    self.trfs = trfs
    if self.trfs is None:
      self.trfs = trf.Compose([trf.ToTensor()])
    
    self.d = d
    self.nc = nc
    self.colors = list(self.cmap.values())
    self.classes = list(self.cmap.keys())

    self.ip = ip if ip[0] == '/' else ip + '/'
    self.lp = lp if lp[0] == '/' else lp + '/'
    
    '''
    imdp = '**/*' if not imdf else '*'
    self.ipf = glob.glob(ip + imdp, recursive=True)
    self.lpf = glob.glob(lp + imdp, recursive=True)
    self.tinp = len(self.ipf)
    '''

    self.bs = bs
    self.len = len(self.inpn) // self.bs
    
    # FIXME Check if the self.get_batch() method works perfectly 
    # compared to pytoch implementation
    self.dltr = self.get_batch()
    '''
    # TODO: Update to use own loaders to support imdf
    ipd = torchvision.datasets.ImageFolder(ip, transform=self.trfs)
    self.dltr = DataLoader(ipd, batch_size=bs, shuffle=True, num_workers=2)
    lpd = torchvision.datasets.ImageFolder(lp, transform=self.trfs)
    self.dltt = DataLoader(lpd, batch_size=bs, shuffle=True, num_workers=2)
    '''
    
    # Check for unusualities in the given directory
    #self.check()

  def show_batch(self, t='fin'):
    """
    Arguments:
    :: t - ['inp', 'lab', 'com']
       inp - The inputs
       lab - The Labels
       com = Inputs and Labels combined

    """
    # Implicitly checks for self.y is not None
    assert self.x is not None
    assert t in ['inp', 'lab', 'fin']

    plt.figure(figsize=(10, 10))
    gs = gridspec.GridSpec(2, 1)
    gs.update(wspace=0.5, hspace=0.5)

    for i in range(self.bs if self.bs <= 10 else 10):
      ax = plt.subplot(gs[i])
      plt.axis('off')
      inp = self.t2n(self.x[i])
      lab = self.decode(self.t2n(self.y[i], c=False))
      fin = cv2.addWeighted(inp, 0.5, lab, 0.5, 0, dtype=0)
      ax.imshow(locals()[t])

    plt.show()
  

  def t2n(self, t, c=True):
    if c:
      return t.transpose(0, 1).transpose(1, 2).detach().cpu().numpy()
    else:
      return t.detach().cpu().numpy()
      

  def one_batch(self):
    self.x, _ = next(iter(self.dltr))
    self.y, _ = next(iter(self.dltt))
    return self.x, self.y
    
  def check(self):
    if self.dataset_is is None and self.color_map is None:
      raise RuntimeError('Both \'dataset_is\' and \'color_map\' can\'t be None')
    super(SegmentationDatasetLoader, self).check()
    

  def create_masks(self, image=None, path=None):
    '''
    A class that creates masks for each of the classes

    Arguments:
    - Image.PIL - Semantic Segmented Image where each pixel is colored
           with a specific color
           The Image is of size H x W x C
           where H is the height of the image
              W is the width of the image
              C is the number of channels (3)

    Returns:
    - np.ndarray - of size N x H x W
           where N is the number of classes
              H is the height of the image
              W is the width of the image
    '''
    
    if image is None and path is None:
      raise RuntimeError('Either image or path needs to be passed!')

    if not path is None:
      if not os.path.exists(path):
        raise RuntimeError('You need to pass a valid path!\n \
                  Try passing a number if you are having trouble reaching \
                  the filename')
      image = np.array(Image.open(path)).astype(np.uint8)

    if image.shape[-1] > 3 or image.shape[-1] < 3:
      raise RuntimeError('The image passed has more than expected channels!')
    
    masks = []
    for ii in self.color_map:
      color_img = []
      for j in range(3):
        color_img.append(np.ones((img.shape[:-1])) * ii[j])
      img2 = np.array(color_img, dtype=np.uint8).transpose(1, 2, 0)
      masks.append(np.uint8((image == img2).sum(axis=-1) == 3))

    return np.array(masks)


  def decode(self, image=None):
    '''
    The method helps one get a colorful image where each color corresponds to each class

    Arguments:
    - Image - np.array - A 2D Image where each pixel position is a number indicating
              the class to which is belongs
    
    Returns:
    - np.array - H x W x C
          where each pixel position [x, y, :]
          is a color representing its RGB color which is passed in
          with the color_map while initializing this class
    '''

    if image is None and path is None:
      raise RuntimeError('Either image or path needs to be passed!')

    r = np.zeros_like(image).astype(np.uint8)
    g = np.zeros_like(image).astype(np.uint8)
    b = np.zeros_like(image).astype(np.uint8)

    for label in range(0, self.nc):
      r[image == label] = self.colors[label][0]
      g[image == label] = self.colors[label][1]
      b[image == label] = self.colors[label][2]
    
    rgb = np.stack([r, g, b], axis=2)
    return rgb

  def __len__(self):
    return self.len

  def get_dltr(self):
    return self.dltr

  def get_dltt(self):
    return self.dltt

def camvid(root:str='.', download:bool=False, **kwargs):
  root = root + '/' if root[-1] is not '/' else root
  
  if download and not os.path.isdir(root + 'camvid/'):
    dirname = os.path.dirname(__file__)
    subprocess.run(['sh', dirname + '/get_datasets/get_camvid.sh', root])

  kwargs['ip'] = root + 'camvid/images/_images'
  kwargs['lp'] = root + 'camvid/labels/_labels'
  kwargs['d']  = 'camvid'
  kwargs['nc'] = 32

  assert os.path.isdir(kwargs['ip']) == True
  assert os.path.isdir(kwargs['ip']) == True

  return SegLoader(**kwargs)



import torch
import torch.nn as nn

from .bnconv import bnconv

class Resblock(nn.Module):
 
 def __init__(self, ic:int, oc:int, stride:int=1, num_conv:int=2, 
       expansion:int=4, conv_first:bool=True, relu_after_add:bool=True, 
       eps:float=1e-5, momentum:float=0.1, inplace=True):
  '''
  This class defines the Residual Basic Block with n Conv Layers

  Arguments;
  - ic :: # of input channels
  - oc :: # of output channels for the final conv layers
  - downsample :: Whether this block is to downsample the input
  - expansion :: The expansion for the channels
         Default: 4
  - stride :: if downsample is True then the specified stride is used.
       Default: 2
  - conv_first :: Whether to apply conv before bn or otherwise.
         Default: True
  
  - num_conv :: # of conv layers in the block for the main branch
  '''
  
  super().__init__()
  
  assert(relu_after_add == True or relu_after_add == False)

  self.relu_after_add = relu_after_add
  self.downsample = True if stride > 1 or ic != oc else False
  oc_convi = oc // expansion
  
  layers = [bnconv(in_channels=ic,
            out_channels=oc_convi,
            kernel_size=3,
            padding=1,
            stride=stride,
            eps=eps,
            momentum=momentum,
            conv_first=conv_first,
            inplace=inplace)]
  
  for _ in range(num_conv-2):
    layers += [bnconv(in_channels=oc_convi,
              out_channels=oc_convi,
              kernel_size=3,
              padding=1,
              stride=1,
              eps=eps,
              momentum=momentum,
              conv_first=conv_first,
              inplace=inplace)]
  
  layers += [bnconv(in_channels=oc_convi,
            out_channels=oc,
            kernel_size=3,
            padding=1,
            stride=1,
            eps=eps,
            momentum=momentum,
            conv_first=conv_first,
            inplace=inplace,
            act=False)]
  
  self.main = nn.Sequential(*layers)

  if self.downsample:
    self.side = bnconv(in_channels=ic,
              out_channels=oc,
              kernel_size=3,
              padding=1,
              stride=stride,
              eps=eps,
              momentum=momentum,
              conv_first=conv_first,
              inplace=inplace,
              act=False)
  
  if self.relu_after_add:
    self.relu = nn.ReLU(inplace=inplace)
  
 def forward(self, x):

  x = self.main(x) + (self.side(x) if self.downsample else x)
  return self.relu(x) if self.relu_after_add else x

class Resnext(nn.Module):
 
 def __init__(self, ic:int, oc:int, stride:int=1, g:int=32, num_conv:int=2, 
       expansion:int=4, conv_first:bool=True, relu_after_add:bool=False, eps:float=1e-5, momentum:float=0.1, inplace=True, icf:int=1):
  '''
  This class defines the Residual Basic Block with n Conv Layers

  Arguments;
  :: ic         - # of input channels
  :: oc         - # of output channels for the final conv layers
  :: downsample - Whether this block is to downsample the input
  :: expansion  - The expansion for the channels
          Default: 4
  :: stride     - if downsample is True then the specified stride is used.
          Default: 2
  :: conv_first - Whether to apply conv before bn or otherwise.
         Default: True
  
  :: num_conv   - # of conv layers in the block for the main branch
  :: g          - # of groups
  :: icf        - Increase the number of input channels by icf in the 
          first conv.
  '''
  
  super().__init__()
  
  assert(relu_after_add == True or relu_after_add == False)

  self.relu_after_add = relu_after_add
  self.downsample = True if stride > 1 or ic != oc else False
  oc_1 = ic * icf
  
  layers = [bnconv(in_channels=ic,
            out_channels=oc_1,
            kernel_size=1,
            padding=0,
            stride=stride,
            eps=eps,
            momentum=momentum,
            conv_first=conv_first,
            inplace=inplace)]
  
  layers += [nn.Conv2d(in_channels=oc_1,
            out_channels=oc_1,
            kernel_size=3,
            padding=1,
            stride=1,
            groups=g,
            bias=False)]
  
  layers += [bnconv(in_channels=oc_1,
            out_channels=oc,
            kernel_size=1,
            padding=0,
            stride=1,
            eps=eps,
            momentum=momentum,
            conv_first=conv_first,
            inplace=inplace,
            act=False)]
  
  self.main = nn.Sequential(*layers)

  if self.downsample:
    self.side = bnconv(in_channels=ic,
              out_channels=oc,
              kernel_size=3,
              padding=1,
              stride=stride,
              eps=eps,
              momentum=momentum,
              conv_first=conv_first,
              inplace=inplace,
              act=False)
  
  if self.relu_after_add:
    self.relu = nn.ReLU(inplace=inplace)
  
 def forward(self, x):
  
  x1 = self.main(x)
  print (x1.shape)
  x = self.main(x) + (self.side(x) if self.downsample else x)
  return self.relu(x) if self.relu_after_add else x

def res_stage(block:nn.Module, ic:int, oc:int, num_layers:int,
       stride:int=2, conv_first=False, inplace=False, lconv:int=2):
  
  '''

  Arguments:
  
  :: block - the block type to be stacked one upon another
  :: ic   - # of input channels
  :: oc  - # of output channels

  :: num_layers - # of blocks to be stacked
  :: stride     - Denotes the stride of the first conv block in the resnet block
          and the stride of the conv block in the side branch.
          If stride > 1 then downsample is triggered.
          Else the HxW is same before and after input and output.
  :: conv_first - Whether
          conv -> norm -> relu (True)
          relu -> norm -> conv (False)
  :: inplace    - Whether to perform inplace operations
  :: lconv      - # of conv in the main branch (resnet) of each stacked block
  '''

  layers = []
  layers += nn.ModuleList([block(ic=ic, oc=oc, stride=stride,
                  conv_first=conv_first, inplace=inplace,
                  num_conv=lconv)])

  layers.extend([block(ic=oc, oc=oc, conv_first=conv_first, 
            num_conv=lconv, inplace=inplace) for i in range (num_layers-1)])
  
  return nn.Sequential(*layers)



import torch
import torch.nn as nn
import torch.optim as optim
import numpy as np

from tqdm import tqdm
from torch.utils.data import DataLoader
from torchvision import datasets, transforms

from scratchai.learners.metrics import accuracy
from scratchai.utils import AvgMeter


def clf_test(net, vloader, crit:nn.Module=nn.CrossEntropyLoss, topk=(1,5)):
  """
  This function helps in quickly testing the network.

  Arguments
  ---------
  net : nn.Module
        The net which to train.
  vloader : torch.nn.utils.DataLoader
            or a generator which returns the images and the labels
  
  """
  # TODO Fix this
  if topk != (1, 5):
    raise Exception('topk other than (1, 5) not supported for now.')

  device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
  a1mtr = AvgMeter('test_acc1'); a5mtr = AvgMeter('test_acc5') #
  vloss = 0
  net.to(device)
  net.eval()
  try: crit = crit()
  except: pass
  with torch.no_grad():
    for ii, (data, labl) in enumerate(tqdm(vloader)):
      data, labl = data.to(device), labl.to(device)
      out = net(data)
      vloss += crit(out, labl).item()
      acc1, acc5 = accuracy(out, labl, topk=topk)
      a1mtr(acc1, data.size(0)); a5mtr(acc5, data.size(0))
    
    vloss /= len(vloader)
  return (a1mtr.avg, a5mtr.avg), vloss


def clf_train(net, tloader, opti:torch.optim, crit:nn.Module, **kwargs):
  # TODO Fix this
  if kwargs['topk'] != (1, 5):
    raise Exception('topk other than (1, 5) not supported for now.')

  device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
  net.to(device)
  net.train()
  a1mtr = AvgMeter('train_acc1'); a5mtr = AvgMeter('train_acc5')
  tloss = 0
  try: crit = crit()
  except: pass
  for ii, (data, labl) in enumerate(tqdm(tloader)):
    data, labl = data.to(device), labl.to(device)
    out = net(data)
    loss = crit(out, labl)
    opti.zero_grad()
    loss.backward()
    opti.step()
    with torch.no_grad():
      tloss += loss.item()
      acc1, acc5 = accuracy(out, labl, topk=kwargs['topk'])
      a1mtr(acc1, data.size(0)); a5mtr(acc5, data.size(0))

  tloss /= len(tloader)
  return (a1mtr.avg, a5mtr.avg), tloss


def clf_fit(net:nn.Module, crit:nn.Module, opti:torch.optim, tloader, vloader, 
            **kwargs):
  """
  This function is used to train the classification networks.
  """
  epochs = kwargs['epochs']
  lr = kwargs['lr']
  lr_step = kwargs['lr_step']
  lr_decay = kwargs['lr_decay']
  seed = kwargs['seed'] if kwargs['seed'] else np.random.randint(100)

  bloss = float('inf')
  
  torch.manual_seed(seed)
  np.random.seed(seed)
  print ('[INFO] Setting torch seed to {}'.format(seed))

  device = 'cuda' if torch.cuda.is_available() else 'cpu'
  
  tlist = []
  vlist = []

  for e in range(1, epochs+1):
    if lr_step is not None and type(lr_step) == int and e%lr_step == 0:
      lr = adjust_lr(opti, lr, lr_decay)
    if lr_step is not None and type(lr_step) == list and e in lr_step:
      lr = adjust_lr(opti, lr, lr_decay)
      
    tacc, tloss = clf_train(net, tloader, opti, crit, topk=kwargs['topk'])
    vacc, vloss = clf_test(net, vloader, crit, topk=kwargs['topk'])
    
    tlist.append((tacc, tloss))
    vlist.append((vacc, vloss))

    if vloss < bloss:
      bloss = vloss
      torch.save({'net' : net.state_dict(), 'opti' : opti.state_dict()},
               'best_net-{}-{:.2f}.pth'.format(e, vacc[0]))
    
    # TODO The tloss and vloss needs a recheck.
    print ('Epoch: {}/{} - Train Loss: {:.3f} - Train Acc@1: {:.3f}' 
           '- Train Acc@5: {:.3f} - Val Loss: {:.3f} - Val Acc@1: {:.3f}'
           '- Val Acc@5: {:.3f}'.format(e, epochs, tloss, tacc[0], tacc[1], 
           vloss, vacc[0], vacc[1]))

    torch.save({'net' : net.cpu().state_dict(), 'opti' : opti.state_dict()},
               'net-{}-{:.2f}.pth'.format(e, vacc[0]))

  return tlist, vlist


def adjust_lr(opti, lr, lr_decay):
  """
  Sets learning rate to the initial LR decayed by 10 every `step` epochs.

  Arguments
  ---------
  opti : torch.optim
         The optimizer
  epoch : int
          The number of epochs completed.
  lr : float
       The initial learning rate.
  step : int
         The lr_step, at what interval lr needs to be updated.
  """
  # See: https://discuss.pytorch.org/t/adaptive-learning-rate/320/4
  lr /= (1. / lr_decay)
  for pgroup in opti.param_groups: pgroup['lr'] = lr
  print ('[INFO] Learning rate decreased to {}'.format(lr))
  return lr



"""
The LeNet.
Paper: http://yann.lecun.com/exdb/publis/pdf/lecun-01a.pdf
"""

import torch
import torch.nn as nn

from scratchai.nets.common import Flatten
from scratchai.utils import load_from_pth
from scratchai.pretrained import urls

__all__ = ['Lenet', 'lenet_mnist', 'lenet_cifar10']

def conv(ic:int, oc:int, k:int=5):
  layers = [nn.Conv2d(ic, oc, kernel_size=5), nn.MaxPool2d(2), 
            nn.ReLU(inplace=True)]
  return layers

class Lenet(nn.Module):
  """
  Implements the Lenet module.
  Paper: http://yann.lecun.com/exdb/publis/pdf/lecun-01a.pdf

  Arguments
  ---------
  nc : int
       # of classes.
  ic : int
       # of input channels
  ex : int
       Expansion factor
  inhw : int
         The height and width of the input image
         (assuming both are same)
  """
  def __init__(self, nc:int=10, ic:int=3, ex:int=1, inhw:int=32):
    super().__init__()
    # TODO Needs refactoring, as its hard coded for 32x32x3 and 28x28x1
    # Change the op a/c to the inhw
    op = 5 if inhw == 32 else 4 # else assumes input size is 28 (for MNIST)
    layers = [*conv(ic, 6*ex), *conv(6*ex, 16*ex), Flatten(), 
              nn.Linear(16*op*op*ex, 120*ex), nn.Linear(120*ex, 84*ex), 
              nn.Linear(84*ex, nc)]
    self.net = nn.Sequential(*layers)

  def forward(self, x): return self.net(x)


def lenet_mnist(pretrained=True, **kwargs):
  kwargs['ic'] = 1; kwargs['inhw'] = 28
  net = Lenet(**kwargs)
  if pretrained: 
    net.load_state_dict(load_from_pth(urls.lenet_mnist_url, 'lenet_mnist'))
  return net

def lenet_cifar10(pretrained=False, **kwargs):
  kwargs['ic'] = 3; kwargs['inhw'] = 32
  net = Lenet(**kwargs)
  if pretrained: 
    net.load_state_dict(load_from_pth(urls.lenet_mnist_url, 'lenet_cifar10'))
  return net



# Imports
import torch
import torch.nn as nn
import numpy as np
import matplotlib.pyplot as plt
import torch.optim as optim
import torch.utils.data.dataloader as dataloader
from torchvision import datasets, transforms
import itertools
import matplotlib.gridspec as gridspec
from tqdm import tqdm_notebook as tqdm
from IPython.display import clear_output
import pydot

def bnconv(in_channels, out_channels, kernel_size, stride, padding, init=nn.init.kaiming_normal_,
           eps=1e-5, momentum=0.1, conv_first=True, relu='relu', nslope=0.2, norm='instance'):
    
    norm = nn.BatchNorm2d if norm == 'batch' else nn.InstanceNorm2d
    acti = nn.ReLU(inplace=True) if relu == 'relu' else nn.LeakyReLU(negative_slope=nslope, inplace=True)
    bias = norm == nn.InstanceNorm2d
    
    conv = nn.Conv2d(in_channels, out_channels, kernel_size=kernel_size, stride=stride, 
                     padding=padding, bias=bias)
    if init:
        init(conv.weight)
        if hasattr(conv, 'bias') and hasattr(conv.bias, 'data'): conv.bias.data.fill_(0.)
            
    layers = [conv, norm(out_channels, eps=eps, momentum=momentum), acti]
    
    return layers if relu is not None else layers[:2]


def bnconvt(in_channels, out_channels, kernel_size, stride, padding, init=nn.init.kaiming_normal_,
            eps=1e-5, momentum=0.1, conv_first=True, relu='relu', nslope=0.2, norm='instance'):

    norm = nn.BatchNorm2d if norm == 'batch' else nn.InstanceNorm2d
    acti = nn.ReLU(inplace=True) if relu == 'relu' else nn.LeakyReLU(negative_slope=nslope, inplace=True)
    bias = norm == nn.InstanceNorm2d
    
    conv = nn.ConvTranspose2d(in_channels, out_channels, kernel_size=kernel_size,
                              stride=stride, padding=1, output_padding=1, bias=bias)
    
    if init:
        init(conv.weight)
        if hasattr(conv, 'bias') and hasattr(conv.bias, 'data'): conv.bias.data.fill_(0.)
        
    layers = [conv, norm(out_channels, eps=eps, momentum=momentum), acti]
    
    return layers if relu is not None else layers[:2]


class ResidualBlock2L(nn.Module):
  
    def __init__(self, ic_conv, oc_conv, expansion=4, stride=2, downsample=False,
                     conv_first=True, relu_after_add=True, norm='instance',
                     pad_type='reflect', eps=1e-5, momentum=0.1):
        '''
        This class defines the Residual Basic Block with 2 Conv Layers

        Arguments;
        - ic_conv : # of input channels
        - oc_conv : # of output channels for the final conv layers
        - downsample : Whether this block is to downsample the input
        - expansion : The expansion for the channels
                      Default: 4
        - stride : if downsample is True then the specified stride is used.
                   Default: 2
        - conv_first : Whether to apply conv before bn or otherwise.
                       Default: True

        - relu_after_add : Whether to apply the relu activation after
                           adding both the main and side branch.
                           Default: True
        '''

        super(ResidualBlock2L, self).__init__()

        assert(downsample == True or downsample == False)
        assert(relu_after_add == True or relu_after_add == False)
        self.downsample = downsample
        self.expansion = expansion
        self.relu_after_add = relu_after_add
        oc_convi = oc_conv // self.expansion

        stride = stride if self.downsample else 1

        layers = []

        p = 0
        if pad_type == 'reflect':
            layers += [nn.ReflectionPad2d(1)]
        elif pad_type == 'replicate':
            layers += [nn.ReplicationPad2d(1)]
        elif pad_type == 'pad':
            p = 1
        else:
            raise NotImplemented('This pad type is not implemented!')

        layers += [*bnconv(in_channels=ic_conv,
                                out_channels=oc_convi,
                                kernel_size=3,
                                padding=p,
                                stride=stride,
                                #eps=2e-5,
                                #momentum=0.9,
                                conv_first=True,
                                norm=norm,
                                eps=eps,
                                momentum=momentum)]

        p = 0
        if pad_type == 'reflect':
            layers += [nn.ReflectionPad2d(1)]
        elif pad_type == 'replicate':
            layers += [nn.ReplicationPad2d(1)]
        elif pad_type == 'pad':
            p = 1
        else:
            raise NotImplemented('This pad type is not implemented!')

        layers += [*bnconv(in_channels=oc_convi,
                            out_channels=oc_conv,
                            kernel_size=3,
                            padding=p,
                            stride=1,
                            #eps=2e-5,
                            #momentum=0.9,
                            conv_first=True,
                            relu=None,
                            norm=norm,
                            eps=eps,
                            momentum=momentum)
                   ]


        self.side = nn.Sequential(*layers)
    
    def forward(self, x): return x + self.side(x)

class CycleGAN_G(nn.Module):

    def __init__(self, ic_conv=64, norm='instance', pad_type='reflect', 
                 eps=1e-5, momentum=0.1):
        '''
        This class defines the network used for CycleGAN.
        This network is solely taken from the Supplementary material provided
        along with the Perceptual Losses paper by Johnson et al.
        '''
        
        super(CycleGAN, self).__init__()
        
        self.net = nn.Sequential(
                        nn.ReflectionPad2d(3),
                        
                        *bnconv(3, ic_conv, kernel_size=7, padding=0, stride=1,
                                conv_first=True, norm=norm, eps=eps, momentum=momentum),

                        *bnconv(ic_conv, ic_conv, kernel_size=3, padding=1, stride=2,
                                conv_first=True, norm=norm, eps=eps, momentum=momentum),

                        *bnconv(ic_conv, ic_conv*2, kernel_size=3, padding=1, stride=2,
                                conv_first=True, norm=norm, eps=eps, momentum=momentum),

                        *bnconv(ic_conv*2, ic_conv*4, kernel_size=3, padding=1, stride=2,
                                conv_first=True, norm=norm, eps=eps, momentum=momentum),

                        ResidualBlock2L(ic_conv*4, ic_conv*4,
                                        conv_first=True,
                                        relu_after_add=False,
                                        norm=norm,
                                        pad_type=pad_type,
                                        eps=eps,
                                        momentum=momentum),

                        ResidualBlock2L(ic_conv*4, ic_conv*4,
                                        conv_first=True,
                                        relu_after_add=False,
                                        norm=norm,
                                        pad_type=pad_type,
                                        eps=eps,
                                        momentum=momentum),

                        ResidualBlock2L(ic_conv*4, ic_conv*4,
                                        conv_first=True,
                                        relu_after_add=False,
                                        norm=norm,
                                        pad_type=pad_type,
                                        eps=eps,
                                        momentum=momentum),

                        ResidualBlock2L(ic_conv*4, ic_conv*4,
                                        conv_first=True,
                                        relu_after_add=False,
                                        norm=norm,
                                        pad_type=pad_type,
                                        eps=eps,
                                        momentum=momentum),

                        ResidualBlock2L(ic_conv*4, ic_conv*4,
                                        conv_first=True,
                                        relu_after_add=False,
                                        norm=norm,
                                        pad_type=pad_type,
                                        eps=eps,
                                        momentum=momentum),
                        
                        ResidualBlock2L(ic_conv*4, oc_conv=ic_conv*4,
                                        conv_first=True,
                                        relu_after_add=False,
                                        norm=norm,
                                        pad_type=pad_type,
                                        eps=eps,
                                        momentum=momentum),

                        *bnconvt(ic_conv*4, ic_conv*2, kernel_size=3, padding=1, stride=2,
                                conv_first=True, norm=norm, eps=eps, momentum=momentum),

                        *bnconvt(ic_conv*2, ic_conv, kernel_size=3, padding=1, stride=2,
                                conv_first=True, norm=norm, eps=eps, momentum=momentum),

                        *bnconvt(ic_conv, ic_conv, kernel_size=3, padding=1, stride=2,
                                conv_first=True, relu=False, norm=norm, eps=eps, momentum=momentum),
                        
                        nn.ReflectionPad2d(3), nn.Conv2d(ic_conv,3, kernel_size=7, padding=0, stride=1),
                        nn.Tanh()
            )

        self.reset_params()

    def forward(self, x): return self.net(x)
    
    def reset_params(self):
        for layer in self.modules():
            if isinstance(layer, nn.Conv2d):
                nn.init.normal_(layer.weight.data, mean=0.0, std=0.02)
            elif isinstance(layer, nn.BatchNorm2d):
                nn.init.normal_(layer.weight.data, mean=1.0, std=0.02)
                nn.init.constant_(layer.bias.data, 0)
                
class CycleGAN_D(nn.Module):

    def __init__(self, iconv=64, num_conv=4):
        
        super(Discriminator, self).__init__()
        
        layers = [nn.Conv2d(3, iconv, kernel_size=4, stride=2, padding=1), 
                  nn.LeakyReLU(negative_slope=0.2, inplace=True)]

        for _ in range(num_conv-2):
            layers += bnconv(iconv, iconv*2, kernel_size=4, stride=2, padding=1,
                                  relu='leaky', nslope=0.2)
            iconv *= 2

        layers += [*bnconv(iconv, iconv*2, kernel_size=4, stride=1, padding=1, relu='leaky', nslope=0.2),
                   nn.Conv2d(iconv*2, 1, kernel_size=4, stride=1, padding=1)]
        
        self.net = nn.Sequential(*layers)
        #self.reset_params()
        
    def forward(self, x): return self.net(x)
    
    def reset_params(self):
        for layer in self.modules():
            if isinstance(layer, nn.Conv2d):
                nn.init.normal_(layer.weight.data, mean=0.0, std=0.02)
            elif isinstance(layer, nn.BatchNorm2d):
                nn.init.normal_(layer.weight.data, mean=1.0, std=0.02)
                nn.init.constant_(layer.bias.data, 0)
                
    def set_requires_grad(self, trfalse):
        for layer in self.parameters():
            layer.requires_grad = trfalse

########################################################################
#################### Image Pool ########################################
import torch
import torch.nn as nn
import numpy as np

class ImagePool(object):
    '''
    This is the class that stores the Pool of Images
    for the updates to make on the Discriminator.

    Arguments:
    - pool_size - The number of images to keep in buffer,
                  And to sample either from the pool or the
                  currently genererated images with a probability 
                  of 0.5
    '''

    def __init__(self, pool_size, device='cpu'):
        
        self.pool_size = pool_size
        self.device = device
        self.pool = []

    def sample(self, images):
        '''
        This function samples the images from the pool
        '''
        
        to_return = []
        for image in images:
            if len(self.pool) < self.pool_size:
                self.pool.append(image.cpu())
                to_return.append(image.unsqueeze(0))
            
            else:
                # p1 is the probability to should the current image
                # to add to return stack or the images in the pool
                p1 = np.random.rand()

                # p2 is probability with which the current image
                # is added to the pool. in a random idx
                p2 = np.random.rand()

                # If the p < 0.5
                # Sample from existing images
                if p1 < 0.5:
                    ridx = np.random.randint(0, self.pool_size)
                    to_return.append(self.pool[ridx].to(device).unsqueeze(0))
                else:
                    to_return.append(image.unsqueeze(0))

                # If the p < 0.6
                # Add the current image to pool
                if p2 < 0.6:
                    ridx = np.random.randint(0, self.pool_size)
                    self.pool[ridx] = image
        
        to_return = torch.cat(to_return, dim=0)
        return to_return


######################################################################
############# Show Results ###########################################
######################################################################

def show(g_x2y, g_y2x, xl, yl):
    x_data, _ = next(iter(xl))
    y_data, _ = next(iter(yl))
    
    with torch.no_grad():
        pred_y = g_x2y(x_data.to(device))
        pred_x = g_y2x(y_data.to(device))
    
    # Prep for showing
    x_data = x_data[0].squeeze(0).transpose(0, 1).transpose(1, 2).detach().cpu().numpy()
    y_data = y_data[0].squeeze(0).transpose(0, 1).transpose(1, 2).detach().cpu().numpy()
    pred_y = pred_y[0].squeeze(0).detach().transpose(0, 1).transpose(1, 2).cpu().numpy()
    pred_x = pred_x[0].squeeze(0).detach().transpose(0, 1).transpose(1, 2).cpu().numpy()
    to_show = [x_data, pred_y, y_data, pred_x]

    # Now show
    plt.figure(figsize = (10, 10))
    gs = gridspec.GridSpec(2, 2)
    gs.update(wspace=0.025, hspace=0.05)

    for i in range(4):
        ax = plt.subplot(gs[i])
        plt.axis('off')
        ax.imshow(unnormalize(to_show[i]))
    
    plt.show()
    
    
def unnormalize(img):
    #return (img * [0.229, 0.224, 0.225]) + [0.485, 0.456, 0.406]
    return (img * [0.5, 0.5, 0.5]) + [0.5, 0.5, 0.5]
    #return img

############################################################
############################################################
############################################################

# NOTE! Not Tested
class CycleGAN(object):
    
    def __init__(self, x_path, y_path, bs=4, transform=None, device='cuda', l_type='mse' betas=(0.5, 0.999), 
                 dlr=0.0002, glr=0.0002, lambA=10, lambB = 10, idt_loss=True, lamb_idt=0.1):
        
        super().__init__()
        
        # Store some vars
        self.d = device
        self.lambA = lambA
        self.lambB = lambB
        self.lamb_idt = lamb_idt
        self.idt_loss = idt_loss

        # Create the models
        self.gx2y = CycleGAN_G().to(self.d)
        self.gy2x = CycleGAN_G().to(self.d)
        self.dx = CycleGAN_D().to(self.d)
        self.dy = CycleGAN_D().to(self.d)
        
        # Create the Dataloaders
        x_data = datasets.ImageFolder(x_path, transform=transform)
        self.xloader = dataloader.DataLoader(x_data, batch_size=bs, shuffle=True)
        y_data = datasets.ImageFolder(y_path, transform=transform)
        self.yloader = dataloader.DataLoader(y_data, batch_size=bs, shuffle=True)

        # Define the criterion
        if l_type == 'mse':
            self.criterion = nn.MSELoss()
        elif l_type == 'l1':
            self.criterion = nn.L1Loss()
            print ('l1')
        self.cycle_c = nn.L1Loss()

        # Define the optimizer
        self.opt_G = optim.Adam(itertools.chain(self.gx2y.parameters(), self.gy2x.parameters()), lr=dlr, betas=betas)
        self.opt_D = optim.Adam(itertools.chain(self.dy.parameters(), self.dx.parameters()), lr=glr, betas=betas)

        # Intialize vars
        self.idt_x = 0
        self.idt_y = 0
    
    def crit(self, out, tar):
        target = torch.ones_like(out) if tar else torch.zeros_like(out)
        return self.criterion(real, target)

    def forward(self):
        self.fakeY = self.gx2y(self.realX)
        self.recnX = self.gy2x(self.fakeY)
        self.fakeX = self.gy2x(self.realY)
        self.recnY = self.gx2y(self.fakeX)

    def backward_G(self):
        self.gly = self.crit(self.fakeY, True)
        self.glx = self.crit(self.fakeX, True)

        self.cx = self.cycle_c(self.recnX, self.realX) * self.lambA
        self.cy = self.cycle_c(self.recnY, self.realY) * self.lambB

        if self.idt_loss:
            self.idt_x = self.criterion(self.gx2y(self.realY), self.realY) * self.lambA * self.lamb_idt
            self.idt_y = self.criterion(self.gy2x(self.realX), self.realX) * self.lambB * self.lamb_idt

        self.g_loss = self.cx + self.cy + self.gly + self.glx + self.idt_x + self.idt_y
        self.g_loss.backward()
    
    def backward_D(self):
        self.loss_fy = self.crit(self.dy(self.fakeY.detach()), False)
        self.loss_fx = self.crit(self.dx(self.fakeX.detach()), False)

        self.loss_ry = self.crit(self.dy(self.realY), True)
        self.loss_rx = self.crit(self.dx(self.realX), True)

        self.dx_loss = (loss_fx + loss_rx) * 0.5
        self.dx_loss.backward()

        self.dy_loss = (loss_fy + loss_fy) * 0.5
        self.dy_loss.backward()

    def do_one_iter(self):
        self.realX, _ = next(iter(self.xloader))
        self.realY, _ = next(iter(self.yloader))
        self.realX, self.realY = self.realX.to(self.d), self.realY.to(self.d)

        self.forward()

        self.dx.set_requires_grad(False); self.dy.set_requires_grad(False)
        self.gx2y.zero_grad(); self.gy2x.zero_grad()
        self.backward_G()
        self.opt_G.step()

        self.dx.set_requires_grad(True); self.dy.set_requires_grad(True)
        self.dx.zero_grad(); self.dy.zero_grad()
        self.backward_D()
        self.opt_D.step()



import torch
import torch.nn as nn
import numpy as np

from tqdm import tqdm
from torchvision import transforms as T
from torchvision import datasets
from torch.utils.data import DataLoader
from scratchai.utils import freeze, Topk
from scratchai.learners.metrics import accuracy
from scratchai.attacks.attacks import *
from scratchai.imgutils import get_trf
from scratchai.utils import name_from_object
from scratchai._config import CIFAR10, MNIST, IMGNET12


def optimize_linear(grads, eps, ordr):
  """
  Solves for optimal input to a linear function under a norm constraint.

  Arguments
  ---------
  grads : torch.Tensor
           The gradients of the input.
  eps : float
        Scalar specifying the constraint region.
  ordr : [np.inf, 1, 2] 
         Order of norm constraint.

  Returns
  -------
  opt_pert : torch.Tensor
             Optimal Perturbation.
  """

  red_ind = list(range(1, len(grads.size())))
  azdiv = torch.tensor(1e-12, dtype=grads.dtype, device=grads.device)

  if ordr == np.inf:
    opt_pert = torch.sign(grads)

  elif ordr == 1:
    abs_grad = torch.abs(grads)
    sign = torch.sign(grads)
    ori_shape = [1] * len(grads.size())
    ori_shape[0] = grads.size(0)

    max_abs_grad, _ = torch.max(abs_grad.view(grads.size(0), -1), 1)
    max_mask = abs_grad.eq(max_abs_grad.view(ori_shape)).float()
    num_ties = max_mask
    for red_scalar in red_ind:
      num_ties = torch.sum(num_ties, red_scalar, keepdims=True)
    opt_pert = sign * max_mask / num_ties
    # TODO tests

  elif ordr == 2:
    # TODO
    square = torch.max(azdiv, torch.sum(grads ** 2, red_ind, keepdim=True))
    opt_pert = grads / torch.sqrt(square)
    # TODO tests
  else:
    raise NotImplementedError('Only L-inf, L1 and L2 norms are '
                              'currently implemented.')

  scaled_pert = eps * opt_pert
  return scaled_pert


def clip_eta(eta, ord, eps):
  """
  Helper fucntion to clip the perturbation to epsilon norm ball.

  Args:
      eta: A tensor with the current perturbation
      ord: Order of the norm (mimics Numpy)
           Possible values: np.inf, 1 or 2.
      eps: Epsilon, bound of the perturbation.
  """

  # Clipping perturbation eta to self.ord norm ball
  if ord not in [np.inf, 1, 2]:
    raise ValueError('ord must be np.inf, 1, or 2.')

  reduce_ind = list(range(1, len(eta.shape)))
  azdiv = torch.tensor(1e-12)

  if ord == np.inf:
    eta = torch.clamp(eta, -eps, eps)
  else:
    if ord == 1:
      raise NotImplementedError("The expression below is not the correct way"
                                    " to project onto the L1 norm ball.")
      norm = torch.max(azdiv, torch.mean(torch.abs(eta), reduce_ind))

    elif ord == 2:
      # azdiv(avoid_zero_div) must go inside sqrt to avoid a divide by zero
      # in the gradient through this operation.
      norm = torch.sqrt(torch.max(azdiv, torch.mean(eta**2, reduce_ind)))

    # We must clip to within the norm ball, not 'normalize' onto the
    # surface of the ball
    factor = min(1., eps / norm)
    eta *= factor

  return eta



##################################################################
######### Functions to help benchmark attacks ####################
##################################################################

def benchmark_atk(atk, net:nn.Module, **kwargs):
  """
  Helper function to benchmark using a particular attack
  on a particular dataset. All benchmarks that are present in this
  repository are created using this function.

  Arguments
  ---------
  atk : scratchai.attacks.attacks
        The attack on which to use.
  net : nn.Module
        The net which is to be attacked.
  root : str
         The root directory of the dataset.
  dfunc : function
          The function that can take the root and torchvision.transforms
          and return a torchvision.Datasets object
          Defaults to datasets.ImageFolder
  trf : torchvision.Transforms
        The transforms that you want to apply.
        Defaults to (get_trf('rz256_cc224_tt_normimgnet')
  bs : int
       The batch size. Defaults to 4.
  
  """
  
  loader, topk, kwargs = pre_benchmark_atk(**kwargs)

  freeze(net)
  print ('[INFO] Net Frozen!')
  atk = atk(net, **kwargs)
  atk_name = name_from_object(atk)
  net_name = name_from_object(net)

  device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
  loss = 0; adv_loss = 0
  oatopk = Topk('original accuracy', topk)
  aatopk = Topk('adversarial accuracy', topk)
  net.to(device); net.eval()
  crit = nn.CrossEntropyLoss()

  for ii, (data, labl) in enumerate(tqdm(loader)):
    adv_data, data = atk(data.to(device).clone()), data.to(device)
    labl = labl.to(device)
    adv_out = net(adv_data); out = net(data)
    loss += crit(out, labl).item()
    adv_loss += crit(adv_out, labl).item()
    acc = accuracy(out, labl, topk)
    adv_acc = accuracy(adv_out, labl, topk)
    oatopk.update(acc, data.size(0)); aatopk.update(adv_acc, data.size(0))
  loss /= len(loader)
  adv_loss /= len(loader)
  
  print ('\nAttack Summary on {} with {} attack:'.format(net_name, atk_name))
  print ('-'*45)
  print (oatopk)
  print ('-'*35)
  print (aatopk)


def pre_benchmark_atk(**kwargs):
  """
  Helper function that sets all the defaults while performing checks
  for all the options passed before benchmarking attacks.
  """

  # Set the Default options if nothing explicit provided
  def_dict = {
    'bs'       : 4,
    'trf'      : get_trf('rz256_cc224_tt_normimgnet'),
    'dset'     : 'NA',
    'root'     : './',
    'topk'     : (1, 5),
    'dfunc'    : datasets.ImageFolder,
    'download' : True,
  }

  for key, val in def_dict.items(): 
    if key not in kwargs: kwargs[key] = val
  
  if kwargs['dset'] == 'NA':
    if 'loader' not in kwargs:
      dset = kwargs['dfunc'](kwargs['root'], transform=kwargs['trf'])
      loader = DataLoader(dset, batch_size=kwargs['bs'], num_workers=2)
    else:
      loader = kwargs['loader']
  
  # Set dataset specific functions here
  else:
    if kwargs['dset'] == IMGNET12:
      dset = datasets.ImageNet(kwargs['root'], split='test',
                      download=kwargs['download'], transform=kwargs['trf'])
    elif kwargs['dset'] == MNIST:
      kwargs['trf'] = get_trf('tt_normmnist')
      kwargs['dfunc'] = datasets.MNIST
      dset = kwargs['dfunc'](kwargs['root'], train=False, 
                     download=kwargs['download'], transform=kwargs['trf'])
    else: raise

    loader = DataLoader(dset, shuffle=False, batch_size=kwargs['bs'])
  topk = kwargs['topk']
    
  for key, val in kwargs.items():
    print ('[INFO] Setting {} to {}.'.format(key, kwargs[key]))

  # Deleting keys that is used just for benchmark_atk() function is 
  # important as the same kwargs dict is passed to initialize the attack
  # So, otherwise the attack will throw an exception
  for key in def_dict:
    del kwargs[key]
  if 'loader' in kwargs: del kwargs['loader']

  return loader, topk, kwargs



"""
Contains functions that helps in weight initialization.
"""

import torch
import torch.nn as nn
import numpy as np

def xavier_normal(m:nn.Module):
  """
  Xavier Normal Initialization to all the conv layers
  And the weight of batch norm is initialized to 1
  and the bias of the batch norm to 0

  Arguments
  ---------
  m : nn.Module
        The net which to init.
  """

  if isinstance(m, nn.Conv2d):
    nn.init.xavier_normal_(m.weight)
    if m.bias is not None:
      nn.init.zeros_(m.bias)
  # Add other norms (like nn.GroupNorm2d)
  elif isinstance(m, nn.BatchNorm2d):
    nn.init.constant_(m.weight, 1)
    nn.init.constant_(m.bias, 0)


def xavier_uniform(m:nn.Module):
  """
  Xavier Uniform Initialization to all the conv layers
  And the weight of batch norm is initialized to 1
  and the bias of the batch norm to 0

  Arguments
  ---------
  m : nn.Module
        The net which to init.
  """

  if isinstance(m, nn.Conv2d):
    nn.init.xavier_uniform_(m.weight)
    if m.bias is not None:
      nn.init.zeros_(m.bias)
  # Add other norms (like nn.GroupNorm2d)
  elif isinstance(m, nn.BatchNorm2d):
    nn.init.constant_(m.weight, 1)
    nn.init.constant_(m.bias, 0)


def kaiming_normal(m:nn.Module):
  """
  Kaiming Normal Initialization to all the conv layers
  And the weight of batch norm is initialized to 1
  and the bias of the batch norm to 0

  Arguments
  ---------
  m : nn.Module
        The net which to init.
  """

  if isinstance(m, nn.Conv2d):
    nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
    if m.bias is not None:
      nn.init.zeros_(m.bias)
  # Add other norms (like nn.GroupNorm2d)
  elif isinstance(m, nn.BatchNorm2d):
    nn.init.constant_(m.weight, 1)
    nn.init.constant_(m.bias, 0)


def kaiming_uniform(m:nn.Module):
  """
  Kaiming Uniform Initialization to all the conv layers
  And the weight of batch norm is initialized to 1
  and the bias of the batch norm to 0

  Arguments
  ---------
  m : nn.Module
        The net which to init.
  """

  if isinstance(m, nn.Conv2d):
    nn.init.kaiming_uniform_(m.weight, mode='fan_out', nonlinearity='relu')
    if m.bias is not None:
      nn.init.zeros_(m.bias)
  elif isinstance(m, nn.Linear):
    nn.init.zeros_(m.bias)
  # Add other norms (like nn.GroupNorm2d)
  elif isinstance(m, nn.BatchNorm2d):
    nn.init.constant_(m.weight, 1)
    nn.init.constant_(m.bias, 0)


def msr_init(m:nn.Module):
  """
  MSR Initialization to all the conv layers
  And the weight of batch norm is initialized to 1
  and the bias of the batch norm to 0

  Arguments
  ---------
  m : nn.Module
        The net which to init.
  """

  if isinstance(m, nn.Conv2d):
    n = m.kernel_size[0] * m.kernel_size[1] * m.in_channels
    nn.init.normal_(m.weight, mean=0, std=np.sqrt(2/n))
    if m.bias is not None:
      nn.init.zeros_(m.bias)
  elif isinstance(m, nn.Linear):
    nn.init.zeros_(m.bias)
  # Add other norms (like nn.GroupNorm2d)
  elif isinstance(m, nn.BatchNorm2d):
    nn.init.constant_(m.weight, 1)
    nn.init.constant_(m.bias, 0)



"""
Semantic adversarial Examples
"""

__all__ = ['semantic', 'Semantic']

def semantic(x, center:bool=True, max_val:float=1.):
  """
  Semantic adversarial examples.
  
  https://arxiv.org/abs/1703.06857

  Note: data must either be centered (so that the negative image can be
  made by simple negation) or must be in the interval of [-1, 1]

  Arguments
  ---------
  net : nn.Module, optional
        The model on which to perform the attack.
  center : bool
           If true, assumes data has 0 mean so the negative image is just negation.
           If false, assumes data is in interval [0, max_val]
  max_val : float
            Maximum value allowed in the input data.
  """

  if center:
    return x*-1
  return max_val - x

################################################################
###### Class to initialize this attack
###### mainly for the use with torchvision.transforms

class Semantic():
  def __init__(self, net=None, **kwargs):
    self.kwargs = kwargs
  def __call__(self, x):
    return semantic(x, **self.kwargs)
